"""
pygame-menu
https://github.com/ppizarror/pygame-menu

RANGE SLIDER
Slider bar between one/two numeric ranges.
"""

__all__ = [

    # Class
    'RangeSlider',
    'RangeSliderManager',

    # Input types
    'RangeSliderRangeValueType',
    'RangeSliderValueFormatType',
    'RangeSliderValueType'

]

import math
import pygame
import pygame_menu

from abc import ABC
from pygame_menu.locals import POSITION_NORTH, POSITION_SOUTH
from pygame_menu.font import FontType, assert_font
from pygame_menu.locals import FINGERUP, FINGERDOWN, FINGERMOTION
from pygame_menu.utils import check_key_pressed_valid, assert_color, assert_vector, \
    make_surface, get_finger_pos, assert_position, parse_padding
from pygame_menu.widgets.core.widget import Widget, WidgetTransformationNotImplemented, \
    AbstractWidgetManager

from pygame_menu._types import Any, CallbackType, Union, List, Tuple, Optional, \
    ColorType, NumberType, Tuple2IntType, NumberInstance, ColorInputType, \
    EventVectorType, Vector2NumberType, VectorType, PaddingType, Tuple4IntType, \
    Callable, Dict

RangeSliderRangeValueType = Union[Vector2NumberType, VectorType]
RangeSliderValueFormatType = Callable[[NumberType], str]
RangeSliderValueType = Union[NumberType, Vector2NumberType]


# noinspection PyMissingOrEmptyDocstring
class RangeSlider(Widget):
    """
    Range slider widget. Offers 1 or 2 sliders for defining a unique value or
    a range of numeric ones.

    If the state of the widget changes the ``onchange`` callback is called. The
    state can change by pressing LEFT/RIGHT, or by mouse/touch events.

    .. code-block:: python

        onchange(range_value, **kwargs)

    If pressing return (apply) on the widget:

    .. code-block:: python

        onreturn(range_value, **kwargs)

    .. note::

        RangeSlider only accepts translation transformation.

    :param title: Range slider title
    :param rangeslider_id: RangeSlider ID
    :param default_value: Default range value, can accept a number or a tuple/list of 2 elements (min, max). If a single number is provided the rangeslider only accepts 1 value, if 2 are provided, the range is enabled (2 values)
    :param range_values: Tuple/list of 2 elements of min/max values of the range slider. Also range can accept a list of numbers, in which case the values of the range slider will be discrete. List must be sorted
    :param range_width: Width of the range in px
    :param increment: Increment of the value if using left/right keys; used only if the range values are not discrete
    :param onchange: Callback when changing the value of the range slider
    :param onreturn: Callback when pressing return (apply) on the range slider
    :param onselect: Function when selecting the widget
    :param range_box_color: Color of the range box between the sliders
    :param range_box_color_readonly: Color of the range box if widget in readonly state
    :param range_box_enabled: Enables a range box between two sliders
    :param range_box_height_factor: Height of the range box (factor of the range title height)
    :param range_box_single_slider: Enables range box if there's only 1 slider instead of 2
    :param range_line_color: Color of the range line
    :param range_line_height: Height of the range line in px
    :param range_margin: Range margin on x-axis and y-axis (x, y) from title in px
    :param range_text_value_color: Color of the range values text
    :param range_text_value_enabled: Enables the range values text
    :param range_text_value_font: Font of the ranges value. If ``None`` the same font as the widget is used
    :param range_text_value_font_height: Height factor of the range value font (factor of the range title height)
    :param range_text_value_margin_f: Margin of the range text values (factor of the range title height)
    :param range_text_value_position: Position of the range text values, can be NORTH or SOUTH. See :py:mod:`pygame_menu.locals`
    :param range_text_value_tick_color: Color of the range text value tick
    :param range_text_value_tick_enabled: Range text value tick enabled
    :param range_text_value_tick_hfactor: Height factor of the range text value tick (factor of the range title height)
    :param range_text_value_tick_number: Number of range value text, the values are placed uniformly distributed
    :param range_text_value_tick_thick: Thickness of the range text value tick in px
    :param repeat_keys: Enable key repeat
    :param repeat_keys_initial_ms: Time in milliseconds before keys are repeated when held in milliseconds
    :param repeat_keys_interval_ms: Interval between key press repetition when held in milliseconds
    :param slider_color: Slider color
    :param slider_height_factor: Height of the slider (factor of the range title height)
    :param slider_sel_highlight_color: Color of the selected slider highlight box effect
    :param slider_sel_highlight_enabled: Selected slider is highlighted
    :param slider_sel_highlight_thick: Thickness of the selected slider highlight
    :param slider_selected_color: Selected slider color
    :param slider_text_value_bgcolor: Background color of the value text on each slider
    :param slider_text_value_color: Color of value text on each slider
    :param slider_text_value_enabled: Enables a value text on each slider
    :param slider_text_value_font: Font of the slider value. If ``None`` the same font as the widget is used
    :param slider_text_value_font_height: Height factor of the slider font (factor of the range title height)
    :param slider_text_value_margin_f: Margin of the slider text values (factor of the range title height)
    :param slider_text_value_padding: Padding of the slider text values
    :param slider_text_value_position: Position of the slider text values, can be NORTH or SOUTH. See :py:mod:`pygame_menu.locals`
    :param slider_text_value_triangle: Draws a triangle between slider text value and slider
    :param slider_thickness: Slider thickness in px
    :param slider_vmargin: Vertical margin of the slider (factor of the range title height)
    :param value_format: Function that format the value and returns a string that is used in the range and slider text
    :param args: Optional arguments for callbacks
    :param kwargs: Optional keyword arguments
    """
    _font_range_value: Optional['pygame.font.Font']
    _font_slider_value: Optional['pygame.font.Font']
    _increment: NumberType
    _increment_shift_factor: float
    _keyrepeat: bool
    _keyrepeat_counters: Dict[int, int]
    _keyrepeat_initial_interval_ms: NumberType
    _keyrepeat_interval_ms: NumberType
    _range_box: 'pygame.Surface'
    _range_box_color: ColorType
    _range_box_color_readonly: ColorType
    _range_box_enabled: bool
    _range_box_height: int
    _range_box_height_factor: NumberType
    _range_box_pos: Tuple2IntType
    _range_box_single_slider: bool
    _range_line: 'pygame.Surface'
    _range_line_color: ColorType
    _range_line_height: int
    _range_line_pos: Tuple2IntType
    _range_margin: Tuple2IntType
    _range_pos: Tuple2IntType
    _range_text_value_color: ColorType
    _range_text_value_enabled: bool
    _range_text_value_font: Optional[FontType]
    _range_text_value_font_height: NumberType
    _range_text_value_margin: int
    _range_text_value_margin_factor: NumberType
    _range_text_value_position: str
    _range_text_value_surfaces: List['pygame.Surface']
    _range_text_value_surfaces_pos: List[Tuple2IntType]
    _range_text_value_tick_color: ColorType
    _range_text_value_tick_enabled: bool
    _range_text_value_tick_height: int
    _range_text_value_tick_height_factor: NumberType
    _range_text_value_tick_number: int
    _range_text_value_tick_surfaces: List['pygame.Surface']
    _range_text_value_tick_surfaces_pos: List[Tuple2IntType]
    _range_text_value_tick_thickness: int
    _range_values: RangeSliderRangeValueType
    _range_width: int
    _scrolling: bool  # Slider is scrolling
    _selected_mouse: bool
    _single: bool  # Range single or double
    _slider: List['pygame.Surface']
    _slider_color: ColorType
    _slider_height: int
    _slider_height_factor: NumberType
    _slider_pos: Tuple[Tuple2IntType, Tuple2IntType]
    _slider_selected: Tuple[bool, bool]
    _slider_selected_color: ColorType
    _slider_selected_highlight_color: ColorType
    _slider_selected_highlight_enabled: bool
    _slider_selected_highlight_thickness: int
    _slider_text_value_bgcolor: ColorType
    _slider_text_value_color: ColorType
    _slider_text_value_enabled: bool
    _slider_text_value_font: Optional[FontType]
    _slider_text_value_font_height: NumberType
    _slider_text_value_margin: int
    _slider_text_value_margin_factor: NumberType
    _slider_text_value_padding: Tuple4IntType
    _slider_text_value_position: str
    _slider_text_value_surfaces: List['pygame.Surface']
    _slider_text_value_surfaces_pos: List[Tuple2IntType]
    _slider_text_value_triangle: bool
    _slider_text_value_vmargin: int
    _slider_thickness: int
    _slider_vmargin: NumberType
    _value: List[NumberType]  # Public value of the slider, generated from the hidden
    _value_format: RangeSliderValueFormatType
    _value_hidden: List[NumberType]  # Hidden value of the slider, modified by events

    def __init__(
        self,
        title: Any,
        rangeslider_id: str = '',
        default_value: RangeSliderValueType = 0,
        range_values: RangeSliderRangeValueType = (0, 1),
        range_width: int = 150,
        increment: NumberType = 0.1,
        onchange: CallbackType = None,
        onreturn: CallbackType = None,
        onselect: CallbackType = None,
        range_box_color: ColorInputType = (6, 119, 206, 170),
        range_box_color_readonly: ColorInputType = (200, 200, 200, 170),
        range_box_enabled: bool = True,
        range_box_height_factor: NumberType = 0.45,
        range_box_single_slider: bool = False,
        range_line_color: ColorInputType = (100, 100, 100),
        range_line_height: int = 2,
        range_margin: Tuple2IntType = (25, 0),
        range_text_value_color: ColorInputType = (80, 80, 80),
        range_text_value_enabled: bool = True,
        range_text_value_font: Optional[FontType] = None,
        range_text_value_font_height: NumberType = 0.4,
        range_text_value_margin_f: NumberType = 0.8,
        range_text_value_position: str = POSITION_SOUTH,
        range_text_value_tick_color: ColorInputType = (60, 60, 60),
        range_text_value_tick_enabled: bool = True,
        range_text_value_tick_hfactor: NumberType = 0.35,
        range_text_value_tick_number: int = 2,
        range_text_value_tick_thick: int = 1,
        repeat_keys: bool = True,
        repeat_keys_initial_ms: NumberType = 400,
        repeat_keys_interval_ms: NumberType = 50,
        slider_color: ColorInputType = (120, 120, 120),
        slider_height_factor: NumberType = 0.7,
        slider_sel_highlight_color: ColorInputType = (0, 0, 0),
        slider_sel_highlight_enabled: bool = True,
        slider_sel_highlight_thick: int = 1,
        slider_selected_color: ColorInputType = (180, 180, 180),
        slider_text_value_bgcolor: ColorInputType = (140, 140, 140),
        slider_text_value_color: ColorInputType = (0, 0, 0),
        slider_text_value_enabled: bool = True,
        slider_text_value_font: Optional[FontType] = None,
        slider_text_value_font_height: NumberType = 0.4,
        slider_text_value_margin_f: NumberType = 1,
        slider_text_value_padding: PaddingType = (0, 4),
        slider_text_value_position: str = POSITION_NORTH,
        slider_text_value_triangle: bool = True,
        slider_thickness: int = 15,
        slider_vmargin: NumberType = 0,
        value_format: RangeSliderValueFormatType = lambda x: str(round(x, 3)),
        *args,
        **kwargs
    ) -> None:
        super(RangeSlider, self).__init__(
            args=args,
            kwargs=kwargs,
            onchange=onchange,
            onreturn=onreturn,
            onselect=onselect,
            title=title,
            widget_id=rangeslider_id
        )

        # Check ranges
        assert_vector(range_values, 0)
        assert len(range_values) >= 2, \
            'range values length must be equal or greater than 2'
        if len(range_values) == 2:
            assert range_values[1] > range_values[0], \
                'range values must be increasing and different numeric values'
        else:
            for i in range(len(range_values) - 1):
                assert range_values[i] < range_values[i + 1], \
                    'range values list must be ordered and different'
        assert isinstance(range_width, int)
        assert range_width > 0, 'range width must be greater than zero'

        # Check default value
        if not isinstance(default_value, NumberInstance):
            assert_vector(default_value, 2)
            assert default_value[0] < default_value[1], \
                'default value vector must be ordered and different (min,max)'
            assert default_value[0] >= range_values[0], \
                f'minimum default value ({default_value[0]}) must be equal or ' \
                f'greater than minimum value of the range ({range_values[0]})'
            assert default_value[1] <= range_values[-1], \
                f'maximum default value ({default_value[1]}) must be lower or ' \
                f'equal than maximum value of the range ({range_values[-1]})'
            default_value = tuple(default_value)

        else:
            assert range_values[0] <= default_value <= range_values[-1], \
                f'default value ({default_value}) must be between minimum and maximum' \
                f' of the range values ({range_values[0]}, {range_values[-1]}), that ' \
                f'is, it must satisfy {range_values[0]}<={default_value}<={range_values[-1]}'

        # If range is discrete, check default value within list
        if len(range_values) > 2:
            if not isinstance(default_value, NumberInstance):
                assert default_value[0] in range_values, \
                    f'min default value ({default_value[0]}) must be within range'
                assert default_value[1] in range_values, \
                    f'max default value ({default_value[1]}) must be within range'
            else:
                assert default_value in range_values, \
                    f'default value ({default_value}) must be within range values'

        # Check increment
        assert isinstance(increment, NumberInstance)
        assert increment >= 0, 'increment must be equal or greater than zero'

        # Check fonts
        if range_text_value_font is not None:
            assert_font(range_text_value_font)
        assert isinstance(range_text_value_font_height, NumberInstance)
        assert 0 < range_text_value_font_height, \
            'range text value font height must be greater than zero'
        if slider_text_value_font is not None:
            assert_font(slider_text_value_font)
        assert isinstance(slider_text_value_font_height, NumberInstance)
        assert 0 < slider_text_value_font_height, \
            'slider text value font height must be greater than zero'

        # Check colors
        range_box_color = assert_color(range_box_color)
        range_box_color_readonly = assert_color(range_box_color_readonly)
        range_line_color = assert_color(range_line_color)
        range_text_value_color = assert_color(range_text_value_color)
        range_text_value_tick_color = assert_color(range_text_value_tick_color)
        slider_color = assert_color(slider_color)
        slider_selected_color = assert_color(slider_selected_color)
        slider_sel_highlight_color = assert_color(slider_sel_highlight_color)
        slider_text_value_bgcolor = assert_color(slider_text_value_bgcolor)
        slider_text_value_color = assert_color(slider_text_value_color)

        # Check dimensions and sizes
        assert isinstance(range_box_height_factor, NumberInstance)
        assert 0 < range_box_height_factor, \
            'height factor must be greater than zero'
        assert isinstance(range_text_value_margin_f, NumberInstance)
        assert 0 < range_text_value_margin_f, \
            'height factor must be greater than zero'
        assert isinstance(slider_text_value_margin_f, NumberInstance)
        assert 0 < slider_text_value_margin_f, \
            'height factor must be greater than zero'
        assert isinstance(slider_height_factor, NumberInstance)
        assert 0 < slider_height_factor, \
            'height factor must be greater than zero'
        assert isinstance(slider_vmargin, NumberInstance)
        slider_text_value_padding = parse_padding(slider_text_value_padding)
        assert isinstance(slider_thickness, int)
        assert slider_thickness > 0, \
            'slider thickness must be greater than zero'
        assert isinstance(range_line_height, int)
        assert range_line_height >= 0, \
            'range line height must be equal or greater than zero'
        assert_vector(range_margin, 2, int)
        assert isinstance(range_text_value_tick_number, int)
        if range_text_value_enabled:
            assert range_text_value_tick_number >= 2, \
                'number of range value must be equal or greater than 2'
        assert isinstance(range_text_value_tick_thick, int)
        assert range_text_value_tick_thick >= 1, \
            'range text tick thickness must be equal or greater than 1 px'
        assert isinstance(slider_sel_highlight_thick, int)
        assert slider_sel_highlight_thick >= 1, \
            'selected highlight thickness must be equal or greater than 1 px'

        # Check positions
        assert_position(range_text_value_position)
        assert range_text_value_position in (POSITION_NORTH, POSITION_SOUTH), \
            'range text value position must be north or south'
        assert_position(slider_text_value_position)
        assert slider_text_value_position in (POSITION_NORTH, POSITION_SOUTH), \
            'slider text value position must be north or south'

        # Check boolean
        assert isinstance(range_box_enabled, bool)
        assert isinstance(range_box_single_slider, bool)
        assert isinstance(range_text_value_enabled, bool)
        assert isinstance(range_text_value_enabled, bool)
        assert isinstance(slider_sel_highlight_enabled, bool)
        assert isinstance(slider_text_value_enabled, bool)
        assert isinstance(slider_text_value_triangle, bool)

        # Check the value format function
        assert callable(value_format)
        assert isinstance(value_format(0), str), \
            'value_format must be a function that accepts only 1 argument ' \
            '(value) and must return a string'

        # Check keyrepeat
        assert isinstance(repeat_keys, bool)
        assert isinstance(repeat_keys_initial_ms, NumberInstance)
        assert isinstance(repeat_keys_interval_ms, NumberInstance)
        assert repeat_keys_initial_ms > 0, \
            'repeat keys initial ms cannot be lower or equal than zero'
        assert repeat_keys_interval_ms > 0, \
            'repeat keys interval ms cannot be lower or equal than zero'

        # Single value
        single = isinstance(default_value, NumberInstance)

        # Convert default value to list
        if isinstance(default_value, NumberInstance):
            default_value = [default_value, 0]
        else:
            default_value = [default_value[0], default_value[1]]

        # Store properties
        self._clock = pygame.time.Clock()
        self._default_value = tuple(default_value)
        self._increment = increment
        self._increment_shift_factor = 0.5
        self._keyrepeat = repeat_keys
        self._keyrepeat_counters = {}  # {event.key: (counter_int, event.unicode)} (look for "***")
        self._keyrepeat_initial_interval_ms = repeat_keys_initial_ms
        self._keyrepeat_interval_ms = repeat_keys_interval_ms
        self._range_box_color = range_box_color
        self._range_box_color_readonly = range_box_color_readonly
        self._range_box_enabled = range_box_enabled
        self._range_box_height_factor = range_box_height_factor
        self._range_box_single_slider = range_box_single_slider
        self._range_line_color = range_line_color
        self._range_line_height = range_line_height
        self._range_margin = range_margin
        self._range_pos = (0, 0)
        self._range_text_value_color = range_text_value_color
        self._range_text_value_enabled = range_text_value_enabled
        self._range_text_value_font = range_text_value_font
        self._range_text_value_font_height = range_text_value_font_height
        self._range_text_value_margin = 0
        self._range_text_value_margin_factor = range_text_value_margin_f
        self._range_text_value_position = range_text_value_position
        self._range_text_value_tick_color = range_text_value_tick_color
        self._range_text_value_tick_enabled = range_text_value_tick_enabled
        self._range_text_value_tick_height = 0
        self._range_text_value_tick_height_factor = range_text_value_tick_hfactor
        self._range_text_value_tick_number = range_text_value_tick_number
        self._range_text_value_tick_thickness = range_text_value_tick_thick
        self._range_values = tuple(range_values)
        self._range_width = range_width
        self._scrolling = False
        self._selected_mouse = False
        self._single = single
        self._slider_color = slider_color
        self._slider_height = 0
        self._slider_height_factor = slider_height_factor
        self._slider_selected = (True, False)
        self._slider_selected_color = slider_selected_color
        self._slider_selected_highlight_color = slider_sel_highlight_color
        self._slider_selected_highlight_enabled = slider_sel_highlight_enabled
        self._slider_selected_highlight_thickness = slider_sel_highlight_thick
        self._slider_text_value_bgcolor = slider_text_value_bgcolor
        self._slider_text_value_color = slider_text_value_color
        self._slider_text_value_enabled = slider_text_value_enabled
        self._slider_text_value_font = slider_text_value_font
        self._slider_text_value_font_height = slider_text_value_font_height
        self._slider_text_value_margin_factor = slider_text_value_margin_f
        self._slider_text_value_padding = slider_text_value_padding
        self._slider_text_value_position = slider_text_value_position
        self._slider_text_value_triangle = slider_text_value_triangle
        self._slider_text_value_vmargin = 0
        self._slider_thickness = slider_thickness
        self._slider_vmargin = slider_vmargin
        self._value = default_value
        self._value_format = value_format
        self._value_hidden = default_value.copy()  # Used when dragging mouse on discrete range

    def value_changed(self) -> bool:
        if self._single:
            return self.get_value() != self._default_value[0]
        return self.get_value() != self._default_value

    def reset_value(self) -> 'Widget':
        if self._single:
            self.set_value(self._default_value[0])
        else:
            self.set_value(self._default_value)
        return self

    def set_value(self, value: RangeSliderValueType) -> None:
        if self._single:
            assert isinstance(value, NumberInstance)
            # noinspection PyTypeChecker
            assert self._range_values[0] <= value <= self._range_values[-1], \
                f'value ({value}) must be within range {self._range_values[0]} <=' \
                f' {value} <= {self._range_values[1]}'
            if len(self._range_values) > 2:
                assert value in self._range_values, \
                    'value must be between range values discrete list'
            value = [value, 0]

        else:
            assert_vector(value, 2)
            assert self._range_values[0] <= value[0], \
                'value must be equal or greater than minimum range value'
            assert value[1] <= self._range_values[-1], \
                'value must be lower or equal than maximum range value'
            assert value[0] < value[1], 'value vector must be ordered'
            if len(self._range_values) > 2:
                assert value[0] in self._range_values
                assert value[1] in self._range_values
            value = [value[0], value[1]]

        self._value = value
        self._value_hidden = self._value.copy()
        self._render()

    def scale(self, *args, **kwargs) -> 'RangeSlider':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'RangeSlider':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'RangeSlider':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'RangeSlider':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'RangeSlider':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'RangeSlider':
        raise WidgetTransformationNotImplemented()

    def get_value(self) -> Union[NumberType, Tuple[NumberType, NumberType]]:
        if self._single:
            return self._value[0]
        return self._value[0], self._value[1]

    def _apply_font(self) -> None:
        if self._range_text_value_font is None:
            self._range_text_value_font = self._font_name
        if self._slider_text_value_font is None:
            self._slider_text_value_font = self._font_name

        self._font_range_value = pygame_menu.font.get_font(
            self._range_text_value_font,
            int(self._range_text_value_font_height * self._font_size)
        )
        self._font_slider_value = pygame_menu.font.get_font(
            self._slider_text_value_font,
            int(self._slider_text_value_font_height * self._font_size)
        )

        # Compute the height
        height = self._font_render_string('TEST').get_height()
        self._range_box_height = int(height * self._range_box_height_factor)
        self._slider_height = int(height * self._slider_height_factor)
        self._range_text_value_margin = int(height * self._range_text_value_margin_factor)
        self._range_text_value_tick_height = int(height * self._range_text_value_tick_height_factor)
        self._slider_text_value_margin = int(height * self._slider_text_value_margin_factor)

    def _draw(self, surface: 'pygame.Surface') -> None:
        # Draw title
        surface.blit(self._surface, self._rect.topleft)

        # Draw range line
        surface.blit(self._range_line, (self._range_line_pos[0] + self._rect.x, self._range_line_pos[1] + self._rect.y))

        # Draw range values and ticks
        for i in range(len(self._range_text_value_surfaces)):
            if self._range_text_value_enabled:
                surface.blit(self._range_text_value_surfaces[i],
                             (self._rect.x + self._range_text_value_surfaces_pos[i][0],
                              self._rect.y + self._range_text_value_surfaces_pos[i][1]))
            if self._range_text_value_tick_enabled:
                tick_i = self._range_text_value_tick_surfaces[i]
                surface.blit(tick_i,
                             (self._rect.x + self._range_text_value_tick_surfaces_pos[i][0],
                              self._rect.y + self._range_text_value_tick_surfaces_pos[i][1]))

        # Draw range box
        if self._range_box_enabled and (not self._single or self._range_box_single_slider):
            surface.blit(self._range_box, (self._range_box_pos[0] + self._rect.x, self._range_box_pos[1] + self._rect.y))

        # Draw sliders
        surface.blit(self._slider[0], (self._slider_pos[0][0] + self._rect.x, self._slider_pos[0][1] + self._rect.y))
        if not self._single:
            surface.blit(self._slider[1], (self._slider_pos[1][0] + self._rect.x, self._slider_pos[1][1] + self._rect.y))

        # Draw slider highlighted
        if self._slider_selected[0] and not self.readonly and self.is_selected():
            pygame.draw.rect(
                surface, self._slider_selected_highlight_color,
                self._get_slider_inflate_rect(0), self._slider_selected_highlight_thickness
            )
        if self._slider_selected[1] and not self.readonly and self.is_selected():
            pygame.draw.rect(
                surface, self._slider_selected_highlight_color,
                self._get_slider_inflate_rect(1), self._slider_selected_highlight_thickness
            )

    def draw_after_if_selected(self, surface: Optional['pygame.Surface']) -> 'RangeSlider':
        super(RangeSlider, self).draw_after_if_selected(surface)
        self.last_surface = surface

        # Draw slider value
        if self._slider_text_value_enabled and not self.readonly:
            surface.blit(
                self._slider_text_value_surfaces[0],
                (self._slider_text_value_surfaces_pos[0][0] + self._rect.x,
                 self._slider_text_value_surfaces_pos[0][1] + self._rect.y)
            )
            if not self._single:
                surface.blit(
                    self._slider_text_value_surfaces[1],
                    (self._slider_text_value_surfaces_pos[1][0] + self._rect.x,
                     self._slider_text_value_surfaces_pos[1][1] + self._rect.y)
                )

        return self

    def _get_slider_inflate_rect(
        self,
        pos: int,
        inflate: Optional[Tuple2IntType] = None,
        to_real_position: bool = False,
        to_absolute_position: bool = False,
        real_position_visible: bool = True
    ) -> 'pygame.Rect':
        """
        Return the slider inflate rect.

        :param pos: Which slider 0/1
        :param inflate: Inflate in x, y
        :param to_real_position: Transform the widget rect to real coordinates. Used by events
        :param to_absolute_position: Transform the widget rect to absolute coordinates. Used by events
        :param real_position_visible: Return only the visible width/height if ``to_real_position=True``
        :return: Slider inflated rect
        """
        if inflate is None:
            inflate = (0, 0)
        s = self._slider[pos]
        rect = s.get_rect()

        rect.x += self._slider_pos[pos][0] + self._rect.x
        rect.y += self._slider_pos[pos][1] + self._rect.y
        rect = pygame.Rect(
            int(rect.x - inflate[0] / 2),
            int(rect.y - inflate[1] / 2),
            int(rect.width + inflate[0]),
            int(rect.height + inflate[1])
        )

        if self._scrollarea is not None:
            assert not (to_real_position and to_absolute_position), \
                'real and absolute positions cannot be True at the same time'
            if to_real_position:
                rect = self._scrollarea.to_real_position(rect, visible=real_position_visible)
            elif to_absolute_position:
                rect = self._scrollarea.to_absolute_position(rect)

        return rect

    def _render(self) -> Optional[bool]:
        if not hasattr(self, '_font_range_value'):
            return False

        if not self._render_hash_changed(self._selected, self._title, self._visible, self.readonly,
                                         self._range_values, self._slider_selected, self._value[0],
                                         self._value[1], self._scrolling, self._selected_mouse):
            return True

        # Create basic title
        self._surface = self._render_string(self._title, self.get_font_color_status())
        self._rect.width, self._rect.height = self._surface.get_size()
        self._range_pos = (self._rect.width + self._range_margin[0], int(self._rect.height / 2) + self._range_margin[1])

        # Create slider
        sel_s = self._slider_selected[0] and self._selected and not self.readonly, \
                self._slider_selected[1] and self._selected and not self.readonly
        self._slider = [
            make_surface(self._slider_thickness, self._slider_height,
                         fill_color=(self._slider_color if not self.readonly else self._font_readonly_color)
                         if not sel_s[0] else self._slider_selected_color),
            make_surface(self._slider_thickness, self._slider_height,
                         fill_color=(self._slider_color if not self.readonly else self._font_readonly_color)
                         if not sel_s[1] else self._slider_selected_color)
        ]
        slider_vm = int(self._slider_vmargin * self._rect.height)  # Vertical margin
        if self._single:
            self._slider_pos = (
                (self._range_pos[0] + self._get_pos_range(self._value[0], self._slider[0]),
                 self._range_pos[1] + slider_vm - int(self._slider_height / 2)),
                (0, 0)
            )
        else:
            self._slider_pos = (
                (self._range_pos[0] + self._get_pos_range(self._value[0], self._slider[0]),
                 self._range_pos[1] + slider_vm - int(self._slider_height / 2)),
                (self._range_pos[0] + self._get_pos_range(self._value[1], self._slider[1]),
                 self._range_pos[1] + slider_vm - int(self._slider_height / 2))
            )

        # Create the range line
        self._range_line = make_surface(self._range_width, self._range_line_height, fill_color=self._range_line_color)
        self._range_line_pos = (self._range_pos[0], int(self._range_pos[1] - self._range_line_height / 2))

        # Create the range font surfaces
        range_values: List[NumberType] = []
        if len(self._range_values) == 2:
            d_val = (self._range_values[1] - self._range_values[0]) \
                    / (self._range_text_value_tick_number - 1)
            v_i = self._range_values[0]
            for i in range(self._range_text_value_tick_number):
                range_values.append(v_i)
                v_i += d_val
        else:
            for i in self._range_values:
                range_values.append(i)

        # Create surfaces for each range value text & position + ticks
        self._range_text_value_surfaces = []
        self._range_text_value_surfaces_pos = []
        self._range_text_value_tick_surfaces = []
        self._range_text_value_tick_surfaces_pos = []
        range_value_pos = 1 if self._range_text_value_position == POSITION_SOUTH else -1
        for i in range_values:
            s = self._font_range_value.render(
                self._value_format(i), self._font_antialias,
                self._range_text_value_color if not self.readonly else self._font_readonly_color)
            self._range_text_value_surfaces.append(s)
            s_x = self._range_pos[0] + self._get_pos_range(i, s)
            s_y = self._range_pos[1] + self._range_text_value_margin / 2 * range_value_pos
            self._range_text_value_surfaces_pos.append((int(s_x), int(s_y)))

            # Create the tick
            s_tick = make_surface(self._range_text_value_tick_thickness,
                                  self._range_text_value_tick_height,
                                  fill_color=self._range_text_value_tick_color)
            self._range_text_value_tick_surfaces.append(s_tick)
            t_x = self._range_pos[0] + self._get_pos_range(i)
            t_y = self._range_pos[1] - s_tick.get_height() / 2
            self._range_text_value_tick_surfaces_pos.append((int(t_x), int(t_y)))

        # Computes the height of the ranges value text for modifying the rect box sizing
        range_values_size = 0, 0  # Stores sizing
        if self._range_text_value_enabled:
            s = self._range_text_value_surfaces[0]
            range_values_size = (int(s.get_width() * 0.7), int(self._range_text_value_surfaces_pos[0][1] + s.get_height() * 0.9))

        # Create the range box surface
        if not self._single or self._range_box_single_slider:
            r_pos = 0 if self._single else self._get_pos_range(self._value[0])
            r_width = self._get_distance_between_sliders()
            self._range_box = make_surface(
                max(0, r_width), self._range_box_height,
                fill_color=self._range_box_color if not self.readonly else self._range_box_color_readonly)
            self._range_box_pos = (self._range_pos[0] + r_pos, self._range_pos[1] - int(self._range_box.get_height() / 2))

        # Create the slider values
        self._slider_text_value_surfaces = []
        self._slider_text_value_surfaces_pos = []
        for v in self._value:
            t = self._font_slider_value.render(self._value_format(v), self._font_antialias, self._slider_text_value_color)  # Value text
            st = make_surface(
                t.get_width() + self._slider_text_value_padding[1] + self._slider_text_value_padding[3],
                t.get_height() + self._slider_text_value_padding[0] + self._slider_text_value_padding[2],
                fill_color=self._slider_text_value_bgcolor)
            st.blit(t, (self._slider_text_value_padding[1], self._slider_text_value_padding[0] + self._slider_text_value_vmargin))

            # Create surface that considers st and the triangle
            tri_height = int(self._slider_text_value_margin / 2) - int(self._slider_height / 2)
            if tri_height and self._slider_text_value_triangle:
                st_root = make_surface(st.get_width(), st.get_height() + tri_height)
                st_root.blit(st, (0, 0))
                mid = st.get_width() / 2
                hig = st.get_height()
                w = tri_height / math.sqrt(3)
                tri_poly = ((mid - w, hig), (mid + w, hig), (mid, hig + tri_height))
                pygame.draw.polygon(st_root, self._slider_text_value_bgcolor, tri_poly)
            else:
                st_root = st

            self._slider_text_value_surfaces.append(st_root)
            st_x = self._range_pos[0] + self._get_pos_range(v, st)
            st_y = self._range_pos[1] - int(self._slider_text_value_margin / 2) - st.get_height()
            self._slider_text_value_surfaces_pos.append((st_x, st_y))

        # Update maximum rect height
        self._rect.height = max(self._rect.height, self._slider_height, self._range_line_height, range_values_size[1])
        self._rect.height += self._range_margin[1]
        self._rect.width += self._range_width + self._range_margin[0] + range_values_size[0]

        # Finals
        self.force_menu_surface_update()

    def _get_pos_range(self, value: NumberType, surface: Optional['pygame.Surface'] = None) -> int:
        """
        Return the position of the surface within range slider.

        :param value: Value
        :param surface: Surface or None
        :return: Position in px
        """
        sw = surface.get_width() / 2 if surface is not None else 0
        if len(self._range_values) == 2:
            d = float(value - self._range_values[0]) / (self._range_values[1] - self._range_values[0])
            return int(d * self._range_width - sw)

        # Find the nearest position
        n, t = self._find_nearest_discrete_range(value), len(self._range_values)
        return int((float(n) / (t - 1)) * self._range_width - sw)

    def _get_distance_between_sliders(self) -> int:
        """
        Returns the distance between both sliders.

        :return: Distance in px
        """
        if not self._single:
            return self._get_pos_range(self._value[1]) - self._get_pos_range(self._value[0])
        return self._get_pos_range(self._value[0])

    def _find_nearest_discrete_range(self, value: NumberType) -> int:
        """
        Return the nearest position of value from range discrete list.

        :param value: Value to find
        :return: Position of the list
        """
        n = 0  # Position of the nearest value
        m = math.inf  # Maximum distance
        t = len(self._range_values)  # Number of values
        for j in range(t):
            k = abs(self._range_values[j] - value)
            if k < m:
                m = k  # Update max
                n = j
        return n

    def _update_value(self, delta: NumberType) -> bool:
        """
        Updates the value of the active slider by delta.

        :param delta: Delta value
        :return: True if value changed
        """
        if self.readonly:
            return False

        old_value = self._value.copy()
        old_value_hidden = self._value_hidden.copy()
        slider_idx = 0 if self._slider_selected[0] else 1

        self._value_hidden[slider_idx] += delta
        if self._value_hidden[0] >= self._value_hidden[1] and not self._single:
            self._value_hidden = old_value_hidden
        else:
            self._value_hidden[slider_idx] = max(
                self._range_values[0], min(self._range_values[-1], self._value_hidden[slider_idx]))

        # Update real value
        if len(self._range_values) == 2:
            self._value = self._value_hidden
        else:
            # Find nearest
            val_index_nearest = self._find_nearest_discrete_range(self._value_hidden[slider_idx])
            self._value[slider_idx] = self._range_values[val_index_nearest]
            if self._value[0] >= self._value[1] and not self._single:
                self._value = old_value

        changed = old_value_hidden != self._value_hidden
        if changed:
            self.change()

        return changed

    def _blur(self) -> None:
        self._selected_mouse = False

    def _focus(self) -> None:
        self._selected_mouse = False

    def _left_right(self, event, left: bool) -> bool:
        """
        Process left and right event keys.

        :param event: Event
        :param left: ``True`` if left, right otherwise
        :return: ``True`` if updated
        """
        self._value_hidden = self._value.copy()  # Update hidden to real value
        if event.key not in self._keyrepeat_counters:
            self._keyrepeat_counters[event.key] = 0
        keys_pressed = pygame.key.get_pressed()

        # If not discrete, apply delta as increment
        if len(self._range_values) == 2:
            mod = 1 if not (keys_pressed[pygame.K_LSHIFT] or keys_pressed[pygame.K_RSHIFT]) \
                else self._increment_shift_factor
            if left:
                mod *= -1
            delta = self._increment * mod

        # If discrete, find the current index and subtract +-1, then find the delta
        # as the difference between two states
        else:
            if not self._single:
                slider_idx = 0 if self._slider_selected[0] else 1
            else:
                slider_idx = 0
            current_val_idx = self._find_nearest_discrete_range(self._value[slider_idx])
            new_val_idx = current_val_idx
            if left:
                new_val_idx = max(0, new_val_idx - 1)
            else:
                new_val_idx = min(len(self._range_values) - 1, new_val_idx + 1)
            delta = self._range_values[new_val_idx] - self._range_values[current_val_idx]

        if self._update_value(delta):
            self._sound.play_key_add()
            return True
        return False

    def _test_get_pos_value(self, value: NumberType, dx: int = 0, dy: int = 0) -> Tuple2IntType:
        """
        Return the position of the value in real coordinates, used for testing.

        :param value: Value to get position
        :param dx: Delta for x position in px
        :param dy: Delta for y position in px
        :return: (x, y) position
        """
        rect = self.get_rect(to_real_position=True, apply_padding=False)
        rect.x += self._range_pos[0] + self._get_pos_range(value)
        rect.y += self._range_pos[1]
        return rect.x + dx, rect.y + dy

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)

        self._clock.tick(60)

        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events)
            return False

        # Get time clock
        time_clock = self._clock.get_time()

        updated = False
        events = self._merge_events(events)  # Extend events with custom events

        for event in events:

            if event.type == pygame.KEYDOWN:  # Check key is valid
                if self._ignores_keyboard_nonphysical() and not check_key_pressed_valid(event):
                    continue

            # Check mouse over
            self._check_mouseover(event)

            # Events
            keydown = self._keyboard_enabled and event.type == pygame.KEYDOWN
            joy_hatmotion = self._joystick_enabled and event.type == pygame.JOYHATMOTION
            joy_axismotion = self._joystick_enabled and event.type == pygame.JOYAXISMOTION
            joy_button_down = self._joystick_enabled and event.type == pygame.JOYBUTTONDOWN

            # Left button
            if (
                keydown and self._ctrl.left(event, self) or
                joy_hatmotion and self._ctrl.joy_left(event, self) or
                joy_axismotion and self._ctrl.joy_axis_x_left(event, self)
            ):
                if self._left_right(event, True):
                    return True

            # Right button
            elif (
                keydown and self._ctrl.right(event, self) or
                joy_hatmotion and self._ctrl.joy_right(event, self) or
                joy_axismotion and self._ctrl.joy_axis_x_right(event, self)
            ):
                if self._left_right(event, False):
                    return True

            # Press enter
            elif (
                keydown and self._ctrl.apply(event, self) or
                joy_button_down and self._ctrl.joy_select(event, self)
            ):
                self.apply()
                return True

            # Tab, switch active slider
            elif keydown and self._ctrl.tab(event, self):
                if self._single:
                    continue
                self._slider_selected = (False, True) if self._slider_selected[0] else (True, False)
                return True

            # Releases key
            elif event.type == pygame.KEYUP and self._keyboard_enabled:
                if event.key in self._keyrepeat_counters:
                    del self._keyrepeat_counters[event.key]

            # User clicks the slider rect
            elif (
                event.type == pygame.MOUSEBUTTONDOWN and self._mouse_enabled or
                event.type == FINGERDOWN and self._touchscreen_enabled and self._menu is not None
            ):
                event_pos = get_finger_pos(self._menu, event)

                # Check which slider is clicked
                rect_slider_0 = self._get_slider_inflate_rect(0, to_real_position=True)
                rect_slider_1 = self._get_slider_inflate_rect(1, to_real_position=True)

                rc_1 = rect_slider_0.collidepoint(*event_pos)
                rc_2 = rect_slider_1.collidepoint(*event_pos)

                old_slider_selected = self._slider_selected
                if not self._single:
                    # Check sliders does not collide each other
                    dist_sliders = self._get_distance_between_sliders()
                    sliders_intersect = dist_sliders <= (self._slider_thickness - 1)

                    if rc_1 and not sliders_intersect:
                        self._slider_selected = (True, False)
                    elif rc_2 and not sliders_intersect:
                        self._slider_selected = (False, True)

                if old_slider_selected != self._slider_selected:
                    updated = True
                    self._render()

                # Check if slider is clicked
                self._scrolling = bool(rc_1 or rc_2)
                self._selected_mouse = True

            # User releases the mouse
            elif (
                event.type == pygame.MOUSEBUTTONUP and self._mouse_enabled and event.button in (1, 2, 3) or
                event.type == FINGERUP and self._touchscreen_enabled and self._menu is not None
            ):
                event_pos = get_finger_pos(self._menu, event)

                # If collides and not scroll, update the value to the clicked position
                rect = self.get_rect(to_real_position=True, apply_padding=False)
                if rect.collidepoint(*event_pos) and not self._scrolling and self._selected_mouse:
                    mouse_x, _ = event_pos
                    topleft, _ = rect.topleft
                    topright, _ = rect.topright

                    # Distance from title
                    dist = mouse_x - (topleft + self._range_pos[0])

                    # Current slider active value
                    val = self._value[0] if self._single else self._value[0 if self._slider_selected[0] else 1]
                    val_px = self._get_pos_range(val)

                    delta = (self._range_values[-1] - self._range_values[0]) * (dist - val_px) / self._range_width

                    if delta != 0 and dist >= 0:  # If clicked true value
                        if self._update_value(delta):
                            updated = True

                self._selected_mouse = False

                # Disables scrolling
                if self._scrolling:
                    self._scrolling = False
                    self._value_hidden = self._value.copy()
                    updated = True

            # User scrolls clicked slider
            elif self._scrolling and self._selected_mouse and (
                event.type == pygame.MOUSEMOTION and self._mouse_enabled and hasattr(event, 'rel') or
                event.type == FINGERMOTION and self._touchscreen_enabled and self._menu is not None
            ):
                rel = event.rel[0] if event.type == pygame.MOUSEMOTION else event.dx * 2 * self._menu.get_window_size()[0]
                delta = (self._range_values[-1] - self._range_values[0]) * rel / self._range_width

                # Check mouse position
                mx, my = event.pos if event.type == pygame.MOUSEMOTION else get_finger_pos(self._menu, event)
                rect = self.get_rect(to_real_position=True, apply_padding=False)

                # Compute position of mouse within valid range
                dist_x = mx - (rect.x + self._range_pos[0])
                dist_y = my - rect.y

                # Get current slider pos if not single
                x_max_min = 0, self._range_width
                if not self._single:
                    slider_idx = 0 if self._slider_selected[0] else 1
                    slider_pos = self._get_pos_range(self._value[slider_idx])
                    x_max_min = slider_pos - 1, slider_pos

                # Check slider within rect
                if delta < 0:
                    in_x = -self._slider_height <= dist_x <= x_max_min[1]
                else:
                    in_x = x_max_min[0] <= dist_x <= self._range_width + self._slider_height
                in_y = 0 <= dist_y <= rect.height

                # Checks mouse changed position and within rect position
                if delta != 0 and in_x and in_y:
                    if self._update_value(delta):
                        updated = True

        # Update key counters
        if self._keyrepeat:
            for key in self._keyrepeat_counters:
                self._keyrepeat_counters[key] += time_clock  # Update clock

                # Generate new key events if enough time has passed:
                if self._keyrepeat_counters[key] >= self._keyrepeat_initial_interval_ms:
                    self._keyrepeat_counters[key] = self._keyrepeat_initial_interval_ms - self._keyrepeat_interval_ms
                    self._add_event(pygame.event.Event(pygame.KEYDOWN, key=key))

        return updated


class RangeSliderManager(AbstractWidgetManager, ABC):
    """
    RangeSlider manager.
    """

    def range_slider(
        self,
        title: str,
        default: RangeSliderValueType,
        range_values: RangeSliderRangeValueType,
        increment: Optional[NumberType] = None,
        onchange: CallbackType = None,
        onreturn: CallbackType = None,
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        rangeslider_id: str = '',
        value_format: RangeSliderValueFormatType = lambda x: str(round(x, 3)),
        width: int = 150,
        **kwargs
    ) -> 'pygame_menu.widgets.RangeSlider':
        """
        Add a range slider to the Menu: Offers 1 or 2 sliders for defining a unique
        value or a range of numeric ones.

        If the state of the widget changes the ``onchange`` callback is called. The
        state can change by pressing LEFT/RIGHT, or by mouse/touch events.

        .. code-block:: python

            onchange(range_value, **kwargs)

        If pressing return (apply) on the widget:

        .. code-block:: python

            onreturn(range_value, **kwargs)

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``range_box_color_readonly``      (tuple, list, str, int, :py:class:`pygame.Color`) - Color of the range box if widget in readonly state
            - ``range_box_color``               (tuple, list, str, int, :py:class:`pygame.Color`) - Color of the range box between the sliders
            - ``range_box_enabled``             (bool) - Enables a range box between two sliders
            - ``range_box_height_factor``       (int, float) - Height of the range box (factor of the range title height)
            - ``range_box_single_slider``       (bool) - Enables range box if there's only 1 slider instead of 2
            - ``range_line_color``              (tuple, list, str, int, :py:class:`pygame.Color`) - Color of the range line
            - ``range_line_height``             (int) - Height of the range line in px
            - ``range_margin``                  (tuple, list) - Range margin on x-axis and y-axis (x, y) from title in px
            - ``range_text_value_color``        (tuple, list, str, int, :py:class:`pygame.Color`) - Color of the range values text
            - ``range_text_value_enabled``      (bool) - Enables the range values text
            - ``range_text_value_font_height``  (int, float) - Height factor of the range value font (factor of the range title height)
            - ``range_text_value_font``         (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) - Font of the ranges value. If ``None`` the same font as the widget is used
            - ``range_text_value_margin_f``     (int, float) - Margin of the range text values (factor of the range title height)
            - ``range_text_value_position``     (str) - Position of the range text values, can be NORTH or SOUTH. See :py:mod:`pygame_menu.locals`
            - ``range_text_value_tick_color``   (tuple, list, str, int, :py:class:`pygame.Color`) - Color of the range text value tick
            - ``range_text_value_tick_enabled`` (bool) - Range text value tick enabled
            - ``range_text_value_tick_hfactor`` (bool) Height factor of the range text value tick (factor of the range title height)
            - ``range_text_value_tick_number``  (int) - Number of range value text, the values are placed uniformly distributed
            - ``range_text_value_tick_thick``   (int) - Thickness of the range text value tick in px
            - ``readonly_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode
            - ``readonly_selected_color``       (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode and is selected
            - ``repeat_keys``                   (bool) - Enable key repeat. ``True`` by default
            - ``repeat_keys_initial_ms``        (int) - Time in milliseconds before keys are repeated when held in milliseconds. ``400`` by default
            - ``repeat_keys_interval_ms``       (int) - Interval between key press repetition when held in milliseconds. ``50`` by default
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``slider_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) - Slider color
            - ``slider_height_factor``          (int, float) - Height of the slider (factor of the range title height)
            - ``slider_sel_highlight_color``    (tuple, list, str, int, :py:class:`pygame.Color`) - Color of the selected slider highlight box effect
            - ``slider_sel_highlight_enabled``  (bool) - Selected slider is highlighted
            - ``slider_sel_highlight_thick``    (int) - Thickness of the selected slider highlight
            - ``slider_selected_color``         (tuple, list, str, int, :py:class:`pygame.Color`) - Selected slider color
            - ``slider_text_value_bgcolor``     (tuple, list, str, int, :py:class:`pygame.Color`) - Background color of the value text on each slider
            - ``slider_text_value_color``       (tuple, list, str, int, :py:class:`pygame.Color`) - Color of value text on each slider
            - ``slider_text_value_enabled``     (bool) - Enables a value text on each slider
            - ``slider_text_value_font_height`` (int, float) - Height factor of the slider font (factor of the range title height)
            - ``slider_text_value_font``        (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) - Font of the slider value. If ``None`` the same font as the widget is used
            - ``slider_text_value_margin_f``    (int, float) - Margin of the slider text values (factor of the range title height)
            - ``slider_text_value_padding``     (int, float, tuple, list) – Padding of the slider text values
            - ``slider_text_value_position``    (str) - Position of the slider text values, can be NORTH or SOUTH. See :py:mod:`pygame_menu.locals`
            - ``slider_text_value_triangle``    (bool) - Draws a triangle between slider text value and slider
            - ``slider_thickness``              (int) - Slider thickness in px
            - ``slider_vmargin``                (int, float) - Vertical margin of the slider (factor of the range title height)
            - ``tab_size``                      (int) – Width of a tab character

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param title: Title of the range slider
        :param default: Default range value, can accept a number or a tuple/list of 2 elements (min, max). If a single number is provided the rangeslider only accepts 1 value, if 2 are provided, the range is enabled (2 values)
        :param range_values: Tuple/list of 2 elements of min/max values of the range slider. Also range can accept a list of numbers, in which case the values of the range slider will be discrete. List must be sorted
        :param increment: Increment of the value if using left/right keys; used only if the range values are not discrete
        :param onchange: Callback executed when changing the value of the range slider
        :param onreturn: Callback executed when pressing return (apply) on the range slider
        :param onselect: Callback executed when selecting the widget
        :param rangeslider_id: ID of the range slider
        :param value_format: Function that format the value and returns a string that is used in the range and slider text
        :param width: Width of the range in px
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.RangeSlider`
        """
        assert_vector(range_values, 0)
        if len(range_values) == 2:
            assert isinstance(increment, NumberInstance), \
                'increment must be defined if the range values are not discrete'
        else:
            if increment is None:
                increment = 1

        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        range_margin = kwargs.pop('range_margin', self._theme.widget_box_margin)
        range_line_color = kwargs.pop('range_line_color', self._theme.widget_font_color)
        range_text_value_color = kwargs.pop('range_text_value_color', self._theme.widget_font_color)
        range_text_value_font_height = kwargs.pop('range_text_value_font_height', 0.6)
        range_text_value_tick_hfactor = kwargs.pop('range_text_value_tick_hfactor', 0.5)
        slider_text_value_font_height = kwargs.pop('slider_text_value_font_height', 0.6)

        widget = RangeSlider(
            title=title,
            rangeslider_id=rangeslider_id,
            default_value=default,
            range_values=range_values,
            range_width=width,
            increment=increment,
            onchange=onchange,
            onreturn=onreturn,
            onselect=onselect,
            range_line_color=range_line_color,
            range_margin=range_margin,
            range_text_value_color=range_text_value_color,
            range_text_value_font_height=range_text_value_font_height,
            range_text_value_tick_hfactor=range_text_value_tick_hfactor,
            slider_text_value_font_height=slider_text_value_font_height,
            value_format=value_format,
            **kwargs
        )
        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - UI SOLAR SYSTEM
Advanced example with a fancy solar system.
"""

__all__ = ['main']

import pygame
import pygame.gfxdraw as gfxdraw
import pygame_menu
from pygame_menu.examples import create_example_window
from pygame_menu.examples._resources import SOLAR_SYSTEM_IMG, NEBULA_IMG

import math
import random
from typing import Dict, Union, Optional, List


class Planet(object):
    """
    Planet object.
    """
    button: Optional['pygame_menu.widgets.Button']
    fontsize: int
    image: 'pygame_menu.BaseImage'
    info: str
    name: str
    period: float
    radius: float
    url: str

    def __init__(
        self,
        image: 'pygame_menu.BaseImage',
        info: str,
        url: str,
        radius: float,
        period: float,
        fontsize: Union[int, float]
    ) -> None:
        """
        Create a planet.

        :param image: Planet image
        :param info: Planet info
        :param url: Info url
        :param radius: Rotation radius
        :param period: Rotation period
        :param fontsize: Button text font size
        """
        self.button = None
        self.fontsize = int(fontsize)
        self.image = image
        self.info = info
        self.name = ''
        self.period = period
        self.radius = radius
        self.url = url


# noinspection SpellCheckingInspection
class SolarSystemApp(object):
    """
    Draws a fancy solar system.
    """
    menu: 'pygame_menu.Menu'
    nebulas: List['pygame_menu.BaseImage']
    planets: Dict[str, 'Planet']
    rotation_velocity: float
    stars: List[List[Union[int, float]]]
    surface: 'pygame.Surface'

    def __init__(self) -> None:
        self.surface = create_example_window('Example - Dynamic Widget Update',
                                             (640, 480), flags=pygame.NOFRAME)

        # Create app theme and menu
        theme = pygame_menu.Theme()

        theme.background_color = (0, 0, 0)
        theme.scrollbar_cursor = pygame_menu.locals.CURSOR_HAND
        theme.title_bar_style = pygame_menu.widgets.MENUBAR_STYLE_TITLE_ONLY
        theme.title_close_button_cursor = pygame_menu.locals.CURSOR_HAND
        theme.title_floating = True
        theme.widget_selection_effect = pygame_menu.widgets.NoneSelection()

        # Load the SS image
        base_img = pygame_menu.BaseImage(SOLAR_SYSTEM_IMG, frombase64=True)

        # Create planets
        self.planets = {
            'sun': Planet(
                base_img.copy().crop(1, 1, 237, 238),
                'The Sun is the Solar System\'s star and by far its most massive '
                'component. Its large mass (332,900 Earth masses), which comprises '
                '99.86% of all the mass in the Solar System, produces temperatures '
                'and densities in its core high enough to sustain nuclear fusion '
                'of hydrogen into helium, making it a main-sequence star. This '
                'releases an enormous amount of energy, mostly radiated into space '
                'as electromagnetic radiation peaking in visible light.',
                'https://en.wikipedia.org/wiki/Sun',
                radius=0,
                period=0,
                fontsize=theme.widget_font_size * 1.25
            ),
            'mercury': Planet(
                base_img.copy().crop(239, 16, 50, 50),
                'Mercury (0.4 AU from the Sun) is the closest planet to the Sun '
                'and on average, all seven other planets. The smallest planet in '
                'the Solar System (0.055 Mo), Mercury has no natural satellites. '
                'Besides impact craters, its only known geological features are '
                'lobed ridges or rupes that were probably produced by a period of '
                'contraction early in its history. Mercury\'s very tenuous '
                'atmosphere consists of atoms blasted off its surface by the solar '
                'wind. Its relatively large iron core and thin mantle have not yet '
                'been adequately explained. Hypotheses include that its outer '
                'layers were stripped off by a giant impact, or that it was '
                'prevented from fully accreting by the young Sun\'s energy.',
                'https://en.wikipedia.org/wiki/Mercury_(planet)',
                radius=0.4,
                period=0.24,
                fontsize=theme.widget_font_size * 0.5
            ),
            'venus': Planet(
                base_img.copy().crop(238, 156, 82, 82),
                'Venus (0.7 AU from the Sun) is close in size to Earth (0.815 Mo) '
                'and, like Earth, has a thick silicate mantle around an iron core, '
                'a substantial atmosphere, and evidence of internal geological '
                'activity. It is much drier than Earth, and its atmosphere is '
                'ninety times as dense. Venus has no natural satellites. It is '
                'the hottest planet, with surface temperatures over 400 °C (752 '
                '°F), most likely due to the amount of greenhouse gases in the '
                'atmosphere. No definitive evidence of current geological '
                'activity has been detected on Venus, but it has no magnetic '
                'field that would prevent depletion of its substantial atmosphere, '
                'which suggests that its atmosphere is being replenished by '
                'volcanic eruptions.',
                'https://en.wikipedia.org/wiki/Venus',
                radius=0.7,
                period=0.615,
                fontsize=theme.widget_font_size * 0.6
            ),
            'earth': Planet(
                base_img.copy().crop(441, 148, 89, 89),
                'Earth (1 AU from the Sun) is the largest and densest of the inner '
                'planets, the only one known to have current geological activity, '
                'and the only place where life is known to exist. Its liquid '
                'hydrosphere is unique among the terrestrial planets, and it is '
                'the only planet where plate tectonics has been observed. Earth\'s '
                'atmosphere is radically different from those of the other planets, '
                'having been altered by the presence of life to contain 21% free '
                'oxygen. It has one natural satellite, the Moon, the only large '
                'satellite of a terrestrial planet in the Solar System.',
                'https://en.wikipedia.org/wiki/Earth',
                radius=1,
                period=1,
                fontsize=theme.widget_font_size * 0.85
            ),
            'moon': Planet(
                base_img.copy().crop(247, 86, 64, 64),
                'The Moon is Earth\'s only proper natural satellite. It is one '
                'quarter the diameter of Earth (comparable to the width of '
                'Australia) making it the largest natural satellite in the Solar '
                'System relative to the size of its planet. It is the fifth '
                'largest satellite in the Solar System and is larger than any '
                'dwarf planet. The Moon orbits Earth at an average lunar distance '
                'of 384,400 km (238,900 mi), or 1.28 light-seconds. Its '
                'gravitational influence produces Earth\'s tides and slightly '
                'lengthens Earth\'s day.',
                'https://en.wikipedia.org/wiki/Moon',
                radius=0.35,
                period=0.2,
                fontsize=theme.widget_font_size * 0.5
            ),
            'mars': Planet(
                base_img.copy().crop(535, 170, 69, 69),
                'Mars (1.5 AU from the Sun) is smaller than Earth and Venus (0.107 '
                'Mo). It has an atmosphere of mostly carbon dioxide with a '
                'surface pressure of 6.1 millibars (roughly 0.6% of that of Earth'
                '). Its surface, peppered with vast volcanoes, such as Olympus '
                'Mons, and rift valleys, such as Valles Marineris, shows '
                'geological activity that may have persisted until as recently '
                'as 2 million years ago. Its red colour comes from iron oxide '
                '(rust) in its soil. Mars has two tiny natural satellites (Deimos '
                'and Phobos) thought to be either captured asteroids, or ejected '
                'debris from a massive impact early in Mars\'s history.',
                'https://en.wikipedia.org/wiki/Mars',
                radius=1.25,
                period=1.880,
                fontsize=theme.widget_font_size * 0.95
            ),
            'jupiter': Planet(
                base_img.copy().crop(322, 89, 118, 118),
                'Jupiter (5.2 AU), at 318 Mo, is 2.5 times the mass of all the '
                'other planets put together. It is composed largely of hydrogen '
                'and helium. Jupiter\'s strong internal heat creates semi-permanent '
                'features in its atmosphere, such as cloud bands and the Great '
                'Red Spot. Jupiter has 79 known satellites. The four largest, '
                'Ganymede, Callisto, Io, and Europa, show similarities to the '
                'terrestrial planets, such as volcanism and internal heating. '
                'Ganymede, the largest satellite in the Solar System, is larger '
                'than Mercury.',
                'https://en.wikipedia.org/wiki/Jupiter',
                radius=1.75,
                period=11.862,
                fontsize=theme.widget_font_size * 1.1
            ),
            'uranus': Planet(
                base_img.copy().crop(525, 83, 83, 83),
                'Uranus (19.2 AU), at 14 Mo, is the lightest of the outer '
                'planets. Uniquely among the planets, it orbits the Sun on its '
                'side; its axial tilt is over ninety degrees to the ecliptic. It '
                'has a much colder core than the other giant planets and radiates '
                'very little heat into space. Uranus has 27 known satellites, the '
                'largest ones being Titania, Oberon, Umbriel, Ariel, and Miranda.',
                'https://en.wikipedia.org/wiki/Uranus',
                radius=2,
                period=84.0205,
                fontsize=theme.widget_font_size
            ),
            'neptune': Planet(
                base_img.copy().crop(448, 1, 92, 92),
                'Neptune (30.1 AU), though slightly smaller than Uranus, is more '
                'massive (17 Mo) and hence more dense. It radiates more internal '
                'heat, but not as much as Jupiter or Saturn. Neptune has 14 known '
                'satellites. The largest, Triton, is geologically active, with '
                'geysers of liquid nitrogen. Triton is the only large satellite '
                'with a retrograde orbit. Neptune is accompanied in its orbit by '
                'several minor planets, termed Neptune trojans, that are in 1:1 '
                'resonance with it.',
                'https://en.wikipedia.org/wiki/Neptune',
                radius=2.25,
                period=164.8,
                fontsize=theme.widget_font_size
            ),
        }

        self.menu = pygame_menu.Menu('Solar System', 640, 480,
                                     onclose=pygame_menu.events.EXIT,
                                     theme=theme, mouse_motion_selection=True)

        # Configure planets and add them to the Menu
        for p in self.planets.keys():
            planet: 'Planet' = self.planets[p]

            # Configure planet
            planet.name = str(p).capitalize()

            # Create submenu for given planet
            submenu = pygame_menu.Menu(planet.name + ' Info', 640, 480, theme=theme,
                                       mouse_motion_selection=True, center_content=False)
            submenu_area_deco = submenu.get_scrollarea().get_decorator()
            submenu_area_deco.add_callable(self.draw_universe_background)

            # Add go back button with a background image
            submenu.add.vertical_margin(150)
            go_back = submenu.add.button('Back to Menu', pygame_menu.events.BACK, cursor=pygame_menu.locals.CURSOR_HAND)
            go_back_img = planet.image.copy().resize(150, 150)
            # Get color from figure's center pixel
            go_back_color = go_back_img.get_at((100, 100), ignore_alpha=True)
            go_back.get_decorator().add_baseimage(0, 0, go_back_img, centered=True)
            go_back_selection = pygame_menu.widgets.HighlightSelection(border_width=2)
            go_back.set_selection_effect(go_back_selection.set_color(go_back_color))

            # Description
            submenu.add.vertical_margin(75)
            submenu.add.label('Description', align=pygame_menu.locals.ALIGN_LEFT,
                              font_name=pygame_menu.font.FONT_OPEN_SANS_BOLD,
                              margin=(5, 10))
            label = submenu.add.label(planet.info, max_char=70,
                                      align=pygame_menu.locals.ALIGN_LEFT,
                                      margin=(29, 1), font_size=20,
                                      font_name=pygame_menu.font.FONT_PT_SERIF,
                                      font_color=(255, 255, 255), padding=0)
            for line in label:
                line.set_max_width(565)
            submenu.add.url(planet.url, align=pygame_menu.locals.ALIGN_LEFT,
                            margin=(20, 1), font_size=20,
                            font_name=pygame_menu.font.FONT_PT_SERIF)
            submenu.add.vertical_margin(40)  # Bottom margin

            # Create advanced button
            planet.image.scale(0.35, 0.35)
            button = self.menu.add.button(planet.name, submenu,
                                          font_size=planet.fontsize)
            button.set_cursor(pygame_menu.locals.CURSOR_HAND)
            button.set_float()
            button.get_decorator().add_baseimage(0, 2, planet.image, centered=True)
            button.set_attribute('planet', planet)
            button.add_draw_callback(self.rotate_planet)
            button_selection = pygame_menu.widgets.LeftArrowSelection(arrow_size=(20, 30), blink_ms=1000)
            button.set_selection_effect(button_selection.set_color(go_back_color))

            # Set random times
            button.set_attribute('t', random.random() * 2 * math.pi)

            # Save button reference to object
            planet.button = button

        # Add draw stars as Menu's decoration
        self.menu.get_scrollarea().get_decorator().add_callable(self.draw_universe_background)

        # Set update event
        self.menu.set_onupdate(self.process_events)

        # Initialize stars random colors
        self.stars = []
        for i in range(100):
            self.add_star()

        # Set the nebulas
        nebula = pygame_menu.BaseImage(NEBULA_IMG, frombase64=True,
                                       drawing_mode=pygame_menu.baseimage.IMAGE_MODE_SIMPLE)
        self.nebulas = [
            nebula.copy().set_drawing_offset((150, 400)).scale2x().set_alpha(25),
            nebula.copy().set_drawing_offset((500, 50)).scale(3, 3).rotate(90).set_alpha(50).pick_channels('b'),
            nebula.copy().set_drawing_offset((50, 100)).scale2x().rotate(175).set_alpha(75).pick_channels('r')
        ]
        for nebula in self.nebulas:
            nebula.set_drawing_position(pygame_menu.locals.POSITION_CENTER)
            nebula.checkpoint()  # Because rotation method works from checkpointed surface
            nebula.set_attribute('delta_angle', 0.25 * random.randint(-1, 1) * random.random())

        # Add shooting stars
        self.shooting_stars = []
        for i in range(3):
            self.add_shooting_star()

        # Update values
        self.rotation_velocity = 0.0003

    def add_shooting_star(self) -> None:
        """
        Add a new shooting star to the simulation.
        """
        dx = random.randrange(-25, 25) * random.random()
        dy = random.randrange(-25, 25) * random.random()
        angle = math.atan2(dy, dx)
        self.shooting_stars.append([
            random.randrange(1, self.menu.get_width()),  # x position
            random.randrange(1, self.menu.get_height()),  # y position
            int(dx),
            int(dy),
            angle,
            random.randrange(1, 5),  # speed
            2 * math.pi * random.random()  # initial flickering
        ])

    def add_star(self) -> None:
        """
        Add a new star to the simulation.
        """
        self.stars.append([
            random.randrange(1, self.menu.get_width()),  # x position
            random.randrange(1, self.menu.get_height()),  # y position
            2 * math.pi * random.random()  # initial flickering
        ])

    # noinspection PyUnusedLocal
    def draw_universe_background(self, surface: 'pygame.Surface', *args) -> None:
        """
        Draw stars as background.

        :param surface: Surface to draw
        :param args: Optional arguments
        """
        t = self.menu.get_counter_attribute('t', self.menu.get_clock().get_time() * 0.001)

        # Draw nebulas
        for nebula in self.nebulas:
            nebula.draw(surface)
            nebula.rotate(nebula.get_angle() + nebula.get_attribute('delta_angle'))

        # Draw stars
        for s in self.stars:
            x, y, flicker = s
            c = int(127 * max(0.5, 1 + math.cos(t + flicker)))
            gfxdraw.pixel(surface, x, y, (c, c, c))

        # Draw shooting stars
        for s in self.shooting_stars:
            x, y, dx, dy, theta, speed, flicker = s
            c = int(150 * max(0.1, max(math.sin(0.1 * t + 1.5 * math.pi + flicker),
                                       math.cos(0.1 * t + flicker))))
            x = int(x)
            y = int(y)
            gfxdraw.line(surface, x, y, x + dx, y + dy, (c, c, c))

            # Update velocity + window constraints
            s[0] = (s[0] + speed * math.cos(theta)) % self.menu.get_width()
            s[1] = (s[1] + speed * math.sin(theta)) % self.menu.get_height()

        # This line forces cache update for submenus that call this method
        self.menu.force_surface_cache_update()

    # noinspection PyProtectedMember
    def process_events(self, events: List['pygame.event.Event'], menu: 'pygame_menu.Menu') -> None:
        """
        Process events from user.

        :param events: Events
        :param menu: Menu
        """
        for event in events:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_p:
                    menu._disable_draw = not menu._disable_draw
                elif event.key == pygame.K_q:
                    if not menu._disable_draw:
                        self.rotation_velocity += 5e-5
                elif event.key == pygame.K_e:
                    if not menu._disable_draw:
                        self.rotation_velocity = max(0.0, self.rotation_velocity - 5e-5)
                elif event.key == pygame.K_s:
                    if not menu._disable_draw:
                        self.add_star()
                elif event.key == pygame.K_c:
                    if not menu._disable_draw:
                        self.add_shooting_star()

    def rotate_planet(self, widget: 'pygame_menu.widgets.Widget', menu: 'pygame_menu.Menu') -> None:
        """
        Rotate a planet.

        :param widget: Widget to rotate
        :param menu: Widget's menu
        """
        # Get planet from attributes
        planet: 'Planet' = widget.get_attribute('planet')

        # Update time from attribute
        t = widget.get_attribute('t')
        if planet.period != 0:
            t -= menu.get_clock().get_time() * self.rotation_velocity / planet.period
        widget.set_attribute('t', t)
        radius_factor = 110  # Visual only

        x = planet.radius * math.cos(t) * radius_factor
        y = planet.radius * math.sin(t) * radius_factor

        # Compute position based on radius
        if planet.name == 'Moon':
            xc, yc = self.planets['earth'].button.get_attribute('pos')  # Center of earth
            x += xc + 1
            y += yc - 10

        # Visual offset
        if planet.name != 'Sun':
            y += 15

        widget.translate(x, y)
        widget.set_attribute('pos', (x, y))
        widget.force_menu_surface_cache_update()

    def mainloop(self, test: bool) -> None:
        """
        App mainloop.

        :param test: Test status
        """
        print('Press ESC to close the app')
        print('Press P to pause the planets')
        print('Press E to increase the rotation velocity')
        print('Press Q to decrease the rotation velocity')
        print('Press S to add a new star')
        print('Press C to add a new shooting star')
        self.menu.mainloop(self.surface, disable_loop=test)


def main(test: bool = False) -> 'SolarSystemApp':
    """
    Main function.

    :param test: Indicate function is being tested
    :return: App
    """
    app = SolarSystemApp()
    app.mainloop(test)
    return app


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

SOUND
Sound class.
"""

__all__ = [

    # Main class
    'Sound',

    # Sound types
    'SOUND_TYPE_CLICK_MOUSE',
    'SOUND_TYPE_CLICK_TOUCH',
    'SOUND_TYPE_CLOSE_MENU',
    'SOUND_TYPE_ERROR',
    'SOUND_TYPE_EVENT',
    'SOUND_TYPE_EVENT_ERROR',
    'SOUND_TYPE_KEY_ADDITION',
    'SOUND_TYPE_KEY_DELETION',
    'SOUND_TYPE_OPEN_MENU',
    'SOUND_TYPE_WIDGET_SELECTION',

    # Sound example paths
    'SOUND_EXAMPLE_CLICK_MOUSE',
    'SOUND_EXAMPLE_CLICK_TOUCH',
    'SOUND_EXAMPLE_CLOSE_MENU',
    'SOUND_EXAMPLE_ERROR',
    'SOUND_EXAMPLE_EVENT',
    'SOUND_EXAMPLE_EVENT_ERROR',
    'SOUND_EXAMPLE_KEY_ADD',
    'SOUND_EXAMPLE_KEY_DELETE',
    'SOUND_EXAMPLE_OPEN_MENU',
    'SOUND_EXAMPLE_WIDGET_SELECTION',
    'SOUND_EXAMPLES'

]

from pathlib import Path
import os.path as path
import time

from pygame import error as pygame_error
from pygame import mixer
from pygame import vernum as pygame_version

from pygame_menu._base import Base
from pygame_menu.utils import warn

from pygame_menu._types import NumberType, Dict, Any, Optional, Union, NumberInstance

try:  # pygame<2.0.0 compatibility
    from pygame import AUDIO_ALLOW_CHANNELS_CHANGE, AUDIO_ALLOW_FREQUENCY_CHANGE
except ImportError:
    AUDIO_ALLOW_CHANNELS_CHANGE, AUDIO_ALLOW_FREQUENCY_CHANGE = False, False

# Sound types
SOUND_TYPE_CLICK_MOUSE = '__pygame_menu_sound_click_mouse__'
SOUND_TYPE_CLICK_TOUCH = '__pygame_menu_sound_click_touch__'
SOUND_TYPE_CLOSE_MENU = '__pygame_menu_sound_close_menu__'
SOUND_TYPE_ERROR = '__pygame_menu_sound_error__'
SOUND_TYPE_EVENT = '__pygame_menu_sound_event__'
SOUND_TYPE_EVENT_ERROR = '__pygame_menu_sound_event_error__'
SOUND_TYPE_KEY_ADDITION = '__pygame_menu_sound_key_addition__'
SOUND_TYPE_KEY_DELETION = '__pygame_menu_sound_key_deletion__'
SOUND_TYPE_OPEN_MENU = '__pygame_menu_sound_open_menu__'
SOUND_TYPE_WIDGET_SELECTION = '__pygame_menu_sound_widget_selection__'

SOUND_TYPES = (
    SOUND_TYPE_CLICK_MOUSE,
    SOUND_TYPE_CLICK_TOUCH,
    SOUND_TYPE_CLOSE_MENU,
    SOUND_TYPE_ERROR,
    SOUND_TYPE_EVENT,
    SOUND_TYPE_EVENT_ERROR,
    SOUND_TYPE_KEY_ADDITION,
    SOUND_TYPE_KEY_DELETION,
    SOUND_TYPE_OPEN_MENU,
    SOUND_TYPE_WIDGET_SELECTION
)

# Sound example paths
__sounds_path__ = path.join(path.dirname(path.abspath(__file__)), 'resources', 'sounds', '{0}')

SOUND_EXAMPLE_CLICK_MOUSE = __sounds_path__.format('click_mouse.ogg')
SOUND_EXAMPLE_CLICK_TOUCH = SOUND_EXAMPLE_CLICK_MOUSE
SOUND_EXAMPLE_CLOSE_MENU = __sounds_path__.format('close_menu.ogg')
SOUND_EXAMPLE_ERROR = __sounds_path__.format('error.ogg')
SOUND_EXAMPLE_EVENT = __sounds_path__.format('event.ogg')
SOUND_EXAMPLE_EVENT_ERROR = __sounds_path__.format('event_error.ogg')
SOUND_EXAMPLE_KEY_ADD = __sounds_path__.format('key_add.ogg')
SOUND_EXAMPLE_KEY_DELETE = __sounds_path__.format('key_delete.ogg')
SOUND_EXAMPLE_OPEN_MENU = __sounds_path__.format('open_menu.ogg')
SOUND_EXAMPLE_WIDGET_SELECTION = __sounds_path__.format('widget_selection.ogg')

SOUND_EXAMPLES = (
    SOUND_EXAMPLE_CLICK_MOUSE,
    SOUND_EXAMPLE_CLICK_TOUCH,
    SOUND_EXAMPLE_CLOSE_MENU,
    SOUND_EXAMPLE_ERROR,
    SOUND_EXAMPLE_EVENT,
    SOUND_EXAMPLE_EVENT_ERROR,
    SOUND_EXAMPLE_KEY_ADD,
    SOUND_EXAMPLE_KEY_DELETE,
    SOUND_EXAMPLE_OPEN_MENU,
    SOUND_EXAMPLE_WIDGET_SELECTION
)

# Stores global reference that marks sounds as initialized
SOUND_INITIALIZED = [False, True]


class Sound(Base):
    """
    Sound engine class.
    
    :param allowedchanges: Convert the samples at runtime, only in pygame>=2.0.0
    :param buffer: Buffer size
    :param channels: Number of channels
    :param devicename: Device name
    :param force_init: Force mixer init with new parameters
    :param frequency: Frequency of sounds
    :param size: Size of sample
    :param sound_id: Sound ID
    :param uniquechannel: Force the channel to be unique, this is set at the object creation moment
    :param verbose: Enable/disable verbose mode (warnings/errors)
    """
    _channel: Optional['mixer.Channel']
    _last_play: str
    _last_time: float
    _mixer_configs: Dict[str, Union[bool, int, str]]
    _sound: Dict[str, Dict[str, Any]]
    _uniquechannel: bool

    def __init__(
        self,
        allowedchanges: int = AUDIO_ALLOW_CHANNELS_CHANGE | AUDIO_ALLOW_FREQUENCY_CHANGE,
        buffer: int = 4096,
        channels: int = 2,
        devicename: str = '',
        force_init: bool = False,
        frequency: int = 22050,
        size: int = -16,
        sound_id: str = '',
        uniquechannel: bool = True,
        verbose: bool = True
    ) -> None:
        super(Sound, self).__init__(object_id=sound_id, verbose=verbose)

        assert isinstance(allowedchanges, int)
        assert isinstance(buffer, int)
        assert isinstance(channels, int)
        assert isinstance(devicename, str)
        assert isinstance(force_init, bool)
        assert isinstance(frequency, int)
        assert isinstance(size, int)
        assert isinstance(uniquechannel, bool)

        assert buffer > 0, 'buffer size must be greater than zero'
        assert channels > 0, 'channels must be greater than zero'
        assert frequency > 0, 'frequency must be greater than zero'

        # Check if mixer is init
        mixer_missing = 'MissingModule' in str(type(mixer))
        if mixer_missing:
            if self._verbose:
                warn('pygame mixer module could not be found, NotImplementedError'
                     'has been raised. Sound support is disabled')
            SOUND_INITIALIZED[1] = False

        # Initialize sounds if not initialized
        if not mixer_missing and ((mixer.get_init() is None and not SOUND_INITIALIZED[0]) or force_init):
            # Set sound as initialized globally
            SOUND_INITIALIZED[0] = True

            # Check pygame version
            version_major, _, version_minor = pygame_version

            # noinspection PyBroadException
            try:
                # pygame < 1.9.5
                mixer_kwargs = {
                    'frequency': frequency,
                    'size': size,
                    'channels': channels,
                    'buffer': buffer
                }

                # pygame >= 1.9.5
                if (version_major == 1 and version_minor > 4) or version_major > 1:
                    mixer_kwargs['devicename'] = devicename

                # pygame >= 2.0.0
                if version_major > 1:
                    mixer_kwargs['allowedchanges'] = allowedchanges

                # Call to mixer
                mixer.init(**mixer_kwargs)

            except Exception as e:
                if self._verbose:
                    warn('sound error: ' + str(e))
            except pygame_error as e:
                if self._verbose:
                    warn('sound engine could not be initialized, pygame error: ' + str(e))

        # Store mixer configs
        self._mixer_configs = {
            'allowedchanges': allowedchanges,
            'buffer': buffer,
            'channels': channels,
            'devicename': devicename,
            'frequency': frequency,
            'size': size
        }

        # Channel where a sound is played
        self._channel = None
        self._uniquechannel = uniquechannel

        # Sound dict
        self._sound = {}
        for sound in SOUND_TYPES:
            self._sound[sound] = {}

        # Last played song
        self._last_play = ''
        self._last_time = 0

    def copy(self) -> 'Sound':
        """
        Return a copy of the object.

        :return: Sound copied
        """
        new_sound = Sound(uniquechannel=self._uniquechannel)
        new_sound._channel = self._channel
        for key in self._mixer_configs:
            new_sound._mixer_configs[key] = self._mixer_configs[key]
        for sound_type in self._sound.keys():
            s = self._sound[sound_type]
            if len(s) != 0:
                new_sound.set_sound(
                    sound_type=sound_type,
                    sound_file=s['path'],
                    volume=s['volume'],
                    loops=s['loops'],
                    maxtime=s['maxtime'],
                    fade_ms=s['fade_ms']
                )
        return new_sound

    def __copy__(self) -> 'Sound':
        """
        Copy method.

        :return: Return new sound
        """
        return self.copy()

    def __deepcopy__(self, memodict: Dict) -> 'Sound':
        """
        Deep-copy method.

        :param memodict: Memo dict
        :return: Return new sound
        """
        return self.copy()

    def get_channel(self) -> 'mixer.Channel':
        """
        Return the channel of the sound engine.

        :return: Sound engine channel
        """
        # noinspection PyArgumentList
        channel = mixer.find_channel()  # force only available on pygame v2
        if self._uniquechannel:  # If the channel is unique
            if self._channel is None:  # If the channel has not been set
                self._channel = channel
        else:
            self._channel = channel  # Store the available channel
        return self._channel

    def set_sound(
        self,
        sound_type: str,
        sound_file: Optional[Union[str, 'Path']],
        volume: float = 0.5,
        loops: int = 0,
        maxtime: NumberType = 0,
        fade_ms: NumberType = 0
    ) -> bool:
        """
        Link a sound file to a sound type.

        :param sound_type: Sound type
        :param sound_file: Sound file. If ``None`` disable the given sound type
        :param volume: Volume of the sound, from ``0.0`` to ``1.0``
        :param loops: Loops of the sound
        :param maxtime: Max playing time of the sound
        :param fade_ms: Fading ms
        :return: The status of the sound load, ``True`` if the sound was loaded
        """
        assert isinstance(sound_type, str)
        assert isinstance(sound_file, (str, type(None), Path))
        assert isinstance(volume, NumberInstance)
        assert isinstance(loops, int)
        assert isinstance(maxtime, NumberInstance)
        assert isinstance(fade_ms, NumberInstance)
        assert loops >= 0, 'loops count must be equal or greater than zero'
        assert maxtime >= 0, 'maxtime must be equal or greater than zero'
        assert fade_ms >= 0, 'fade_ms must be equal or greater than zero'
        assert 1 >= volume >= 0, 'volume must be between 0 and 1'

        # Check sound type is correct
        if sound_type not in SOUND_TYPES:
            raise ValueError('sound type not valid, check the manual')

        # If file is none disable the sound
        if sound_file is None or not SOUND_INITIALIZED[1]:
            self._sound[sound_type] = {}
            return False

        # Check the file exists
        sound_file = str(sound_file)
        if not path.isfile(sound_file):
            raise IOError(f'sound file "{sound_file}" does not exist')

        # Load the sound
        try:
            # noinspection PyTypeChecker
            sound_data = mixer.Sound(file=sound_file)
        except pygame_error:
            if self._verbose:
                warn(f'the sound file "{sound_file}" could not be loaded, it has been disabled')
            self._sound[sound_type] = {}
            return False

        # Configure the sound
        sound_data.set_volume(float(volume))

        # Store the sound
        self._sound[sound_type] = {
            'fade_ms': fade_ms,
            'file': sound_data,
            'length': sound_data.get_length(),
            'loops': loops,
            'maxtime': maxtime,
            'path': sound_file,
            'type': sound_type,
            'volume': volume
        }
        return True

    def load_example_sounds(self, volume: float = 0.5) -> 'Sound':
        """
        Load the example sounds provided by the package.

        :param volume: Volume of the sound, from ``0`` to ``1``
        :return: Self reference
        """
        assert isinstance(volume, NumberInstance) and 0 <= volume <= 1
        for sound in range(len(SOUND_TYPES)):
            self.set_sound(SOUND_TYPES[sound], SOUND_EXAMPLES[sound], volume=float(volume))
        return self

    def _play_sound(self, sound: Optional[Dict[str, Any]]) -> bool:
        """
        Play a sound.

        :param sound: Sound to be played
        :return: ``True`` if the sound was played
        """
        if not sound:
            return False

        # Find an available channel
        channel = self.get_channel()  # This will set the channel if it's None
        if channel is None:  # The sound can't be played because all channels are busy
            return False

        # Play the sound
        sound_time = time.time()

        # If the previous sound is the same and has not ended (max 10% overlap)
        if (
            sound['type'] != self._last_play or
            sound_time - self._last_time >= 0.1 * sound['length'] or
            self._uniquechannel
        ):
            try:
                if self._uniquechannel:  # Stop the current channel if it's unique
                    channel.stop()
                channel.play(sound['file'],
                             loops=sound['loops'],
                             maxtime=sound['maxtime'],
                             fade_ms=sound['fade_ms']
                             )
            except pygame_error:  # Ignore errors
                pass

        # Store last execution
        self._last_play = sound['type']
        self._last_time = sound_time
        return True

    def play_click_mouse(self) -> 'Sound':
        """
        Play click mouse sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_CLICK_MOUSE])
        return self

    def play_click_touch(self) -> 'Sound':
        """
        Play click touch sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_CLICK_TOUCH])
        return self

    def play_error(self) -> 'Sound':
        """
        Play error sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_ERROR])
        return self

    def play_event(self) -> 'Sound':
        """
        Play event sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_EVENT])
        return self

    def play_event_error(self) -> 'Sound':
        """
        Play event error sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_EVENT_ERROR])
        return self

    def play_key_add(self) -> 'Sound':
        """
        Play key addition sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_KEY_ADDITION])
        return self

    def play_key_del(self) -> 'Sound':
        """
        Play key deletion sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_KEY_DELETION])
        return self

    def play_open_menu(self) -> 'Sound':
        """
        Play open Menu sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_OPEN_MENU])
        return self

    def play_close_menu(self) -> 'Sound':
        """
        Play close Menu sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_CLOSE_MENU])
        return self

    def play_widget_selection(self) -> 'Sound':
        """
        Play widget selection sound.

        :return: Self reference
        """
        self._play_sound(self._sound[SOUND_TYPE_WIDGET_SELECTION])
        return self

    def stop(self) -> 'Sound':
        """
        Stop the channel.

        :return: Self reference
        """
        channel = self.get_channel()
        if channel is None:  # The sound can't be played because all channels are busy
            return self
        try:
            channel.stop()
        except pygame_error:
            pass
        return self

    def pause(self) -> 'Sound':
        """
        Pause the channel.

        :return: Self reference
        """
        channel = self.get_channel()
        if channel is None:  # The sound can't be played because all channels are busy
            return self
        try:
            channel.pause()
        except pygame_error:
            pass
        return self

    def unpause(self) -> 'Sound':
        """
        Unpause channel.

        :return: Self reference
        """
        channel = self.get_channel()
        if channel is None:  # The sound can't be played because all channels are busy
            return self
        try:
            channel.unpause()
        except pygame_error:
            pass
        return self

    def get_channel_info(self) -> Dict[str, Any]:
        """
        Return the channel information.

        :return: Information dict e.g.: ``{'busy': 0, 'endevent': 0, 'queue': None, 'sound': None, 'volume': 1.0}``
        """
        channel = self.get_channel()
        data = {}
        if channel is None:  # The sound can't be played because all channels are busy
            return data
        data['busy'] = channel.get_busy()
        data['endevent'] = channel.get_endevent()
        data['queue'] = channel.get_queue()
        data['sound'] = channel.get_sound()
        data['volume'] = channel.get_volume()
        return data



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

TOGGLE SWITCH
Switch between several states.
"""

__all__ = [
    'ToggleSwitch',
    'ToggleSwitchManager'
]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.font import FontType, assert_font
from pygame_menu.locals import FINGERUP
from pygame_menu.utils import check_key_pressed_valid, assert_color, assert_vector, \
    make_surface, get_finger_pos
from pygame_menu.widgets.core.widget import Widget, WidgetTransformationNotImplemented, \
    AbstractWidgetManager

from pygame_menu._types import Any, CallbackType, Union, List, Tuple, Optional, \
    ColorType, NumberType, Tuple2NumberType, Tuple2IntType, NumberInstance, \
    ColorInputType, EventVectorType, Callable


# noinspection PyMissingOrEmptyDocstring
class ToggleSwitch(Widget):
    """
    Toggle switch widget.

    If the state of the widget changes the ``onchange`` callback is called. The
    state can change by pressing LEFT/RIGHT or RETURN if the widget only has two
    states. This widget can handle more than 2 states.

    .. code-block:: python

        onchange(state_value, **kwargs)

    .. note::

        ToggleSwitch only accepts translation transformation.

    :param title: Toggle switch title
    :param toggleswitch_id: ToggleSwitch ID
    :param default_state: Default state index of the switch
    :param infinite: The state can rotate
    :param onchange: Callback when changing the state of the switch
    :param onselect: Function when selecting the widget
    :param single_click: Changes the state of the switch with 1 click instead of finding the closest position (events). If ``True`` the parameter ``infinite`` will also be ``True``
    :param single_click_dir: Direction of the change if only 1 click is pressed. ``True`` for left direction, ``False`` for right
    :param slider_color: Slider color
    :param slider_height_factor: Height of the slider (factor of the switch height)
    :param slider_thickness: Slider thickness in px
    :param slider_vmargin: Vertical margin of the slider (factor of the switch height)
    :param state_color: Background color of each state, it modifies the whole width of the switch
    :param state_text: Text of each state of the switch
    :param state_text_font: Font of the state text. If ``None`` uses the widget font
    :param state_text_font_color: Color of the font of each state text
    :param state_text_font_size: Font size of the state text. If ``None`` uses the widget font size
    :param state_text_position: Position of the state text respect to the switch rect
    :param state_values: Value of each state of the switch
    :param state_width: Width of each state. For example if there's 2 states, ``state_width`` only can have 1 value
    :param switch_border_color: Border color of the switch
    :param switch_border_width: Border width of the switch in px
    :param switch_height: Height factor respect to the title font size height
    :param switch_margin: Switch margin on x-axis and y-axis (x, y) respect to the title of the widget in px
    :param args: Optional arguments for callbacks
    :param kwargs: Optional keyword arguments
    """
    _infinite: bool
    _single_click: bool
    _single_click_dir: bool
    _slider: Optional['pygame.Surface']
    _slider_color: ColorType
    _slider_height: int
    _slider_height_factor: float
    _slider_pos: Tuple2IntType
    _slider_thickness: int
    _slider_vmargin: int
    _state: int
    _state_color: Tuple[ColorType, ...]
    _state_font: Optional['pygame.font.Font']
    _state_text: Tuple[str, ...]
    _state_text_font: Optional[FontType]
    _state_text_font_color: Tuple[ColorType, ...]
    _state_text_font_size: Optional[int]
    _state_text_position: Tuple2NumberType
    _state_values: Tuple[Any, ...]
    _state_width: List[int]
    _switch: Optional['pygame.Surface']
    _switch_border_color: ColorType
    _switch_border_width: int
    _switch_font_rendered: List['pygame.Surface']
    _switch_height: int
    _switch_height_factor: float
    _switch_margin: Tuple2IntType
    _switch_pos: Tuple2IntType
    _switch_width: int
    _total_states: int

    def __init__(
        self,
        title: Any,
        toggleswitch_id: str = '',
        default_state: int = 0,
        infinite: bool = False,
        onchange: CallbackType = None,
        onselect: CallbackType = None,
        single_click: bool = False,
        single_click_dir: bool = True,
        slider_color: ColorInputType = (255, 255, 255),
        slider_height_factor: NumberType = 1,
        slider_thickness: int = 25,
        slider_vmargin: NumberType = 0,
        state_color: Tuple[ColorInputType, ...] = ((178, 178, 178), (117, 185, 54)),
        state_text: Tuple[str, ...] = ('Off', 'On'),
        state_text_font: Optional[FontType] = None,
        state_text_font_color: Tuple[ColorInputType, ...] = ((255, 255, 255), (255, 255, 255)),
        state_text_font_size: Optional[int] = None,
        state_text_position: Tuple2NumberType = (0.5, 0.5),
        state_values: Tuple[Any, ...] = (False, True),
        state_width: Union[Tuple[int, ...], int] = 150,
        switch_border_color: ColorInputType = (40, 40, 40),
        switch_border_width: int = 1,
        switch_height: NumberType = 1.25,
        switch_margin: Tuple2IntType = (25, 0),
        *args,
        **kwargs
    ) -> None:
        super(ToggleSwitch, self).__init__(
            args=args,
            kwargs=kwargs,
            onchange=onchange,
            onselect=onselect,
            title=title,
            widget_id=toggleswitch_id
        )

        # Asserts
        assert isinstance(default_state, int)
        assert isinstance(infinite, bool)
        assert isinstance(single_click, bool)
        assert isinstance(single_click_dir, bool)
        assert isinstance(state_values, tuple)

        # Check the number of states
        self._total_states = len(state_values)
        assert 2 <= self._total_states, 'the minimum number of states is 2'
        assert 0 <= default_state < self._total_states, 'invalid default state value'

        # Check fonts
        if state_text_font is not None:
            assert_font(state_text_font)
        assert isinstance(state_text_font_size, (int, type(None)))
        if state_text_font_size is not None:
            assert state_text_font_size > 0, \
                'state text font size must be equal or greater than zero'

        # Check colors
        assert_vector(state_text_position, 2)
        switch_border_color = assert_color(switch_border_color)
        assert isinstance(switch_border_width, int) and switch_border_width >= 0, \
            'border width must be equal or greater than zero'
        slider_color = assert_color(slider_color)

        # Check dimensions and sizes
        assert slider_height_factor > 0, 'slider height factor cannot be negative'
        assert slider_thickness >= 0, 'slider thickness cannot be negative'
        assert isinstance(slider_vmargin, NumberInstance)
        assert_vector(switch_margin, 2, int)
        assert isinstance(switch_height, NumberInstance) and switch_height > 0, \
            'switch height factor cannot be zero or negative'
        assert isinstance(state_color, tuple) and len(state_color) == self._total_states

        # Create state color
        new_state_color = []
        for c in state_color:
            new_state_color.append(assert_color(c))
        state_color = tuple(new_state_color)

        # Create state texts
        assert isinstance(state_text, tuple) and len(state_text) == self._total_states
        for c in state_text:
            assert isinstance(c, str), 'all states text must be string-type'
        assert isinstance(state_text_font_color, tuple) and \
               len(state_text_font_color) == self._total_states

        new_state_text_font_color = []
        for c in state_text_font_color:
            new_state_text_font_color.append(assert_color(c))
        state_text_font_color = tuple(new_state_text_font_color)

        # Crete state widths
        self._switch_width = 0
        if isinstance(state_width, NumberInstance):
            state_width = [state_width]
        assert_vector(state_width, self._total_states - 1, int)

        for i in range(len(state_width)):
            assert isinstance(state_width[i], int), 'each state width must be an integer'
            assert state_width[i] > 0, 'each state width must be greater than zero'
            self._switch_width += state_width[i]

        # Finals
        if single_click:
            infinite = True

        # Store properties
        self._default_value = default_state
        self._switch_border_color = switch_border_color
        self._switch_border_width = switch_border_width
        self._infinite = infinite
        self._single_click = single_click
        self._single_click_dir = single_click_dir
        self._slider_color = slider_color
        self._slider_height_factor = slider_height_factor
        self._slider_thickness = slider_thickness
        self._slider_vmargin = slider_vmargin
        self._state = default_state
        self._state_color = state_color
        self._state_text = state_text
        self._state_text_font = state_text_font
        self._state_text_font_color = state_text_font_color
        self._state_text_font_size = state_text_font_size
        self._state_text_position = state_text_position
        self._state_values = state_values
        self._state_width = state_width
        self._switch_height_factor = float(switch_height)
        self._switch_margin = switch_margin

        # Compute state width accum
        self._state_width_accum = [0]
        accum = 0
        for w in self._state_width:
            accum += w
            accum_width = accum - self._slider_thickness - 2 * self._switch_border_width
            self._state_width_accum.append(accum_width)

        # Inner properties
        self._accept_events = True
        self._slider_height = 0
        self._slider_pos = (0, 0)  # to add to (rect.x, rect.y)
        self._state_font = None
        self._switch_font_rendered = []  # Stores font render for each state
        self._switch_height = 0
        self._switch_pos = (0, 0)  # horizontal pos, and delta to title

    def set_value(self, state: int) -> None:
        assert isinstance(state, int), 'state value can only be an integer'
        assert 0 <= state < self._total_states, 'state value exceeds the total states'
        self._state = state
        self._render()

    def scale(self, *args, **kwargs) -> 'ToggleSwitch':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'ToggleSwitch':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'ToggleSwitch':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'ToggleSwitch':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'ToggleSwitch':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'ToggleSwitch':
        raise WidgetTransformationNotImplemented()

    def get_value(self) -> Any:
        return self._state_values[self._state]

    def _apply_font(self) -> None:
        self._state_font = pygame_menu.font.get_font(
            self._font_name if self._state_text_font is None else self._state_text_font,
            self._font_size if self._state_text_font_size is None else self._state_text_font_size
        )

        # Compute the height
        height = self._font_render_string('TEST').get_height()
        self._switch_height = int(height * self._switch_height_factor)
        self._slider_height = int(self._switch_height * self._slider_height_factor) - 2 * self._switch_border_width

        # Render the state texts
        self._switch_font_rendered = []
        for t in range(self._total_states):
            f_render = self._state_font.render(self._state_text[t], True, self._state_text_font_color[t])
            self._switch_font_rendered.append(f_render)

    def _draw(self, surface: 'pygame.Surface') -> None:
        # Draw title
        surface.blit(self._surface, (self._rect.x, self._rect.y + self._switch_pos[1] - 1))

        # Draw switch
        switch_x = self._rect.x + self._switch_margin[0] + self._switch_pos[0]
        switch_y = self._rect.y + self._switch_margin[1]
        surface.blit(self._switch, (switch_x, switch_y))

        # Draw switch border
        if self._switch_border_width > 0:
            switch_rect = self._switch.get_rect()
            switch_rect.x += switch_x
            switch_rect.y += switch_y
            pygame.draw.rect(
                surface,
                self._switch_border_color,
                switch_rect,
                self._switch_border_width
            )

        # Draw switch font render
        if self._state_text[self._state] != '':
            text = self._switch_font_rendered[self._state]
            surface.blit(text, (
                int(switch_x + (self._switch_width - text.get_width()) * self._state_text_position[0]),
                int(switch_y + (self._switch_height - text.get_height()) * self._state_text_position[1])
            ))

        # Draw slider
        slider_x = switch_x + self._slider_pos[0] + self._switch_border_width
        slider_y = switch_y + self._slider_pos[1] + self._switch_border_width
        surface.blit(self._slider, (slider_x, slider_y))

    def _render(self) -> Optional[bool]:
        if not self._render_hash_changed(self._selected, self._title, self._visible, self.readonly, self._state):
            return True

        # Create basic title
        self._surface = self._render_string(self._title, self.get_font_color_status())
        self._rect.width, self._rect.height = self._surface.get_size()

        # Create slider
        self._slider = make_surface(self._slider_thickness, self._slider_height, fill_color=self._slider_color)
        self._slider_pos = (self._state_width_accum[self._state],
                            self._slider_vmargin * self._switch_height)

        # Create the switch surface
        self._switch = make_surface(self._switch_width, self._switch_height, fill_color=self._state_color[self._state])
        self._switch_pos = (self._rect.width, int((self._switch_height - self._rect.height) / 2))

        # Update maximum rect height
        self._rect.height = max(self._rect.height, self._switch_height, self._slider_height)
        self._rect.width += self._switch_margin[0] + self._switch_width

        # Finals
        self.force_menu_surface_update()

    def _left(self) -> None:
        """
        State left.
        """
        if self.readonly:
            return
        previous = self._state
        if self._infinite:
            self._state = (self._state - 1) % self._total_states
        else:
            self._state = max(0, self._state - 1)
        if previous != self._state:
            self.change()
            self._sound.play_key_add()

    def _right(self) -> None:
        """
        State right.
        """
        if self.readonly:
            return
        previous = self._state
        if self._infinite:
            self._state = (self._state + 1) % self._total_states
        else:
            self._state = min(self._state + 1, self._total_states - 1)
        if previous != self._state:
            self.change()
            self._sound.play_key_add()

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)

        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events)
            return False

        for event in events:

            if event.type == pygame.KEYDOWN:  # Check key is valid
                if self._ignores_keyboard_nonphysical() and not check_key_pressed_valid(event):
                    continue

            # Check mouse over
            self._check_mouseover(event)

            # Events
            keydown = self._keyboard_enabled and event.type == pygame.KEYDOWN
            joy_hatmotion = self._joystick_enabled and event.type == pygame.JOYHATMOTION
            joy_axismotion = self._joystick_enabled and event.type == pygame.JOYAXISMOTION

            # Left button
            if (
                keydown and self._ctrl.left(event, self) or
                joy_hatmotion and self._ctrl.joy_left(event, self) or
                joy_axismotion and self._ctrl.joy_axis_x_left(event, self)
            ):
                self._left()
                return True

            # Right button
            elif (
                keydown and self._ctrl.right(event, self) or
                joy_hatmotion and self._ctrl.joy_right(event, self) or
                joy_axismotion and self._ctrl.joy_axis_x_right(event, self)
            ):
                self._right()
                return True

            # Press enter
            elif (
                keydown and self._ctrl.apply(event, self) and self._total_states == 2 or
                event.type == pygame.JOYBUTTONDOWN and self._joystick_enabled and self._ctrl.joy_select(event, self) and self._total_states == 2
            ):
                self._sound.play_key_add()
                self._state = int(not self._state)
                self.change()
                self.active = not self.active
                return True

            # Click on switch; don't consider the mouse wheel (button 4 & 5)
            elif (
                event.type == pygame.MOUSEBUTTONUP and self._mouse_enabled and event.button in (1, 2, 3) or
                event.type == FINGERUP and self._touchscreen_enabled and self._menu is not None
            ):
                event_pos = get_finger_pos(self._menu, event)

                # If collides
                rect = self.get_rect(to_real_position=True, apply_padding=False)
                if rect.collidepoint(*event_pos):
                    # Check if mouse collides left or right as percentage, use
                    # only X coordinate
                    mouse_x, _ = event_pos
                    topleft, _ = rect.topleft
                    topright, _ = rect.topright
                    # Distance from title
                    dist = mouse_x - (topleft + self._switch_margin[0] + self._switch_pos[0])

                    if dist > 0:  # User clicked the options, not title
                        # Toggle with only 1 click
                        if self._single_click:
                            if self._single_click_dir:
                                self._left()
                            else:
                                self._right()
                            return True

                        else:
                            target_index = 0
                            best = 1e6
                            # Find the closest position
                            for i in range(self._total_states):
                                dx = abs(self._state_width_accum[i] - dist)
                                if dx < best:
                                    target_index = i
                                    best = dx
                            if target_index != self._state:
                                self._sound.play_key_add()
                                self._state = target_index
                                self.change()
                                return True

        return False


class ToggleSwitchManager(AbstractWidgetManager, ABC):
    """
    ToggleSwitch manager.
    """

    def toggle_switch(
        self,
        title: Any,
        default: Union[int, bool] = 0,
        onchange: CallbackType = None,
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        toggleswitch_id: str = '',
        single_click: bool = True,
        state_text: Tuple[str, ...] = ('Off', 'On'),
        state_values: Tuple[Any, ...] = (False, True),
        width: int = 150,
        **kwargs
    ) -> 'pygame_menu.widgets.ToggleSwitch':
        """
        Add a toggle switch to the Menu: It can switch between two states.

        If user changes the status of the callback, ``onchange`` is fired:

        .. code-block:: python

            onchange(current_state_value, **kwargs)

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``infinite``                      (bool) – The state can rotate. ``False`` by default
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``readonly_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode
            - ``readonly_selected_color``       (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode and is selected
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``single_click_dir``              (bool) - Direction of the change if only 1 click is pressed. ``True`` for left direction (default), ``False`` for right
            - ``slider_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the slider
            - ``slider_height_factor``          (int, float) - Height of the slider (factor of the switch height). ``1`` by default
            - ``slider_thickness``              (int) – Slider thickness in px. ``20`` px by default
            - ``slider_vmargin``                (int, float) - Vertical margin of the slider (factor of the switch height). ``0`` by default
            - ``state_color``                   (tuple) – 2-item color tuple for each state
            - ``state_text_font``               (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`, None) - Font of the state text. If ``None`` uses the widget font. ``None`` by default
            - ``state_text_font_color``         (tuple) – 2-item color tuple for each font state text color
            - ``state_text_font_size``          (str, None) – Font size of the state text. If ``None`` uses the widget font size
            - ``state_text_position``           (tuple) - Position of the state text respect to the switch rect. ``(0.5, 0.5)`` by default
            - ``switch_border_color``           (tuple, list, str, int, :py:class:`pygame.Color`) – Switch border color
            - ``switch_border_width``           (int) – Switch border width
            - ``switch_height``                 (int, float) – Height factor respect to the title font size height
            - ``switch_margin``                 (tuple, list) – Switch on x-axis and y-axis (x, y) margin respect to the title of the widget in px
            - ``tab_size``                      (int) – Width of a tab character

        .. note::

            This method only handles two states. If you need more states (for example
            3, or 4), prefer using :py:class:`pygame_menu.widgets.ToggleSwitch`
            and add it as a generic widget.

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param title: Title of the toggle switch
        :param default: Default state index of the switch; it can be ``0 (False)`` or ``1 (True)``
        :param onchange: Callback executed when changing the state of the toggle switch
        :param onselect: Callback executed when selecting the widget
        :param toggleswitch_id: Widget ID
        :param single_click: Changes the state of the switch with 1 click instead of finding the closest position
        :param state_text: Text of each state
        :param state_values: Value of each state of the switch
        :param width: Width of the switch box in px
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.ToggleSwitch`
        """
        if isinstance(default, (int, bool)):
            assert 0 <= default <= 1, 'default value can be 0 or 1'
        else:
            raise ValueError(
                f'invalid value type, default can be 0, False, 1, or True, but'
                f'received "{default}"'
            )

        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        infinite = kwargs.pop('infinite', False)
        slider_color = kwargs.pop('slider_color', self._theme.widget_box_background_color)
        slider_thickness = kwargs.pop('slider_thickness', self._theme.scrollbar_thick)
        state_color = kwargs.pop('state_color', ((178, 178, 178), (117, 185, 54)))
        state_text_font_color = kwargs.pop('state_text_font_color', (self._theme.widget_box_background_color, self._theme.widget_box_background_color))
        state_text_font_size = kwargs.pop('state_text_font_size', None)
        switch_border_color = kwargs.pop('switch_border_color', self._theme.widget_box_border_color)
        switch_border_width = kwargs.pop('switch_border_width', self._theme.widget_box_border_width)
        switch_height = kwargs.pop('switch_height', 1)
        switch_margin = kwargs.pop('switch_margin', self._theme.widget_box_margin)

        widget = ToggleSwitch(
            default_state=default,
            infinite=infinite,
            onchange=onchange,
            onselect=onselect,
            single_click=single_click,
            single_click_dir=kwargs.pop('single_click_dir', True),
            slider_color=slider_color,
            slider_thickness=slider_thickness,
            state_color=state_color,
            state_text=state_text,
            state_text_font_color=state_text_font_color,
            state_text_font_size=state_text_font_size,
            state_values=state_values,
            switch_border_color=switch_border_color,
            switch_border_width=switch_border_width,
            switch_height=switch_height,
            switch_margin=switch_margin,
            title=title,
            state_width=int(width),
            toggleswitch_id=toggleswitch_id,
            **kwargs
        )
        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

BASEIMAGE
Provides a class to perform basic image loading and manipulation with pygame.
"""
# File constants no. 100

__all__ = [

    # Base class
    'BaseImage',

    # Image paths
    'IMAGE_EXAMPLE_CARBON_FIBER',
    'IMAGE_EXAMPLE_GRAY_LINES',
    'IMAGE_EXAMPLE_METAL',
    'IMAGE_EXAMPLE_PYGAME_MENU',
    'IMAGE_EXAMPLE_PYTHON',
    'IMAGE_EXAMPLE_TILED_BORDER',
    'IMAGE_EXAMPLE_WALLPAPER',
    'IMAGE_EXAMPLES',

    # Drawing modes
    'IMAGE_MODE_CENTER',
    'IMAGE_MODE_FILL',
    'IMAGE_MODE_REPEAT_X',
    'IMAGE_MODE_REPEAT_XY',
    'IMAGE_MODE_REPEAT_Y',
    'IMAGE_MODE_SIMPLE'

]

from io import BytesIO
from pathlib import Path
import base64
import math
import os.path as path

import pygame

from pygame_menu._base import Base
from pygame_menu.locals import POSITION_NORTHWEST, POSITION_NORTHEAST, POSITION_CENTER, \
    POSITION_WEST, POSITION_SOUTHWEST, POSITION_EAST, POSITION_SOUTHEAST, \
    POSITION_SOUTH, POSITION_NORTH
from pygame_menu.utils import assert_vector, assert_position, assert_color, \
    load_pygame_image_file

from pygame_menu._types import Tuple2IntType, Union, Vector2NumberType, Callable, \
    Tuple, List, NumberType, Optional, Dict, Tuple4IntType, Literal, Tuple2NumberType, \
    ColorInputType, Tuple3IntType, NumberInstance, VectorInstance

# Example image paths
__images_path__ = path.join(path.dirname(path.abspath(__file__)), 'resources', 'images', '{0}')

IMAGE_EXAMPLE_CARBON_FIBER = __images_path__.format('carbon_fiber.png')
IMAGE_EXAMPLE_GRAY_LINES = __images_path__.format('gray_lines.png')
IMAGE_EXAMPLE_METAL = __images_path__.format('metal.png')
IMAGE_EXAMPLE_PYGAME_MENU = __images_path__.format('pygame_menu.png')
IMAGE_EXAMPLE_PYTHON = __images_path__.format('python.svg')
IMAGE_EXAMPLE_TILED_BORDER = __images_path__.format('tiled_border.png')
IMAGE_EXAMPLE_WALLPAPER = __images_path__.format('wallpaper.jpg')

IMAGE_EXAMPLES = (IMAGE_EXAMPLE_CARBON_FIBER, IMAGE_EXAMPLE_GRAY_LINES,
                  IMAGE_EXAMPLE_METAL, IMAGE_EXAMPLE_PYGAME_MENU, IMAGE_EXAMPLE_PYTHON,
                  IMAGE_EXAMPLE_TILED_BORDER, IMAGE_EXAMPLE_WALLPAPER)

# Drawing modes
IMAGE_MODE_CENTER = 100
IMAGE_MODE_FILL = 101
IMAGE_MODE_REPEAT_X = 102
IMAGE_MODE_REPEAT_XY = 103
IMAGE_MODE_REPEAT_Y = 104
IMAGE_MODE_SIMPLE = 105  # Just draw the image without any effect

_VALID_IMAGE_MODES = (IMAGE_MODE_CENTER, IMAGE_MODE_FILL, IMAGE_MODE_REPEAT_X,
                      IMAGE_MODE_REPEAT_XY, IMAGE_MODE_REPEAT_Y, IMAGE_MODE_SIMPLE)

# Other constants
_VALID_IMAGE_FORMATS = ['.jpg', '.png', '.gif', '.bmp', '.pcx', '.tga', '.tif',
                        '.lbm', '.pbm', '.pgm', '.ppm', '.xpm', '.svg', 'BytesIO',
                        'base64']

# Custom types
ColorChannelType = Literal['r', 'g', 'b']
ChannelType = Union[ColorChannelType, Tuple[ColorChannelType, ColorChannelType], Tuple[ColorChannelType, ColorChannelType, ColorChannelType], List[ColorChannelType]]


class BaseImage(Base):
    """
    Object that loads an image, stores as a surface, transform it and
    let write the image to a surface.

    :param image_path: Path of the image to be loaded. It can be a string (path, base64), :py:class:`pathlib.Path`, or :py:class:`io.BytesIO`
    :param drawing_mode: Drawing mode of the image
    :param drawing_offset: Offset of the image in drawing method
    :param drawing_position: Drawing position if mode is ``IMAGE_MODE_SIMPLE``. See :py:mod:`pygame_menu.locals` for valid ``position`` values
    :param load_from_file: Loads the image from the given path
    :param frombase64: If ``True`` consider ``image_path`` as base64 string
    :param image_id: str
    """
    _angle: NumberType
    _drawing_mode: int
    _drawing_offset: Tuple2IntType
    _drawing_position: str
    _extension: str
    _filename: str
    _filepath: Union[str, 'BytesIO']
    _frombase64: bool
    _last_transform: Tuple[int, int, Optional['pygame.Surface']]
    _original_surface: 'pygame.Surface'
    _rotated: bool
    _surface: 'pygame.Surface'
    smooth_scaling: bool

    def __init__(
        self,
        image_path: Union[str, 'Path', 'BytesIO'],
        drawing_mode: int = IMAGE_MODE_FILL,
        drawing_offset: Vector2NumberType = (0, 0),
        drawing_position: str = POSITION_NORTHWEST,
        load_from_file: bool = True,
        frombase64: bool = False,
        image_id: str = ''
    ) -> None:
        super(BaseImage, self).__init__(object_id=image_id)

        assert isinstance(image_path, (str, Path, BytesIO)), \
            'path must be string, Path, or BytesIO object type'
        assert isinstance(load_from_file, bool)
        assert isinstance(frombase64, bool)

        if isinstance(image_path, (str, Path)):
            image_path = str(image_path)
            if not frombase64:
                _, file_extension = path.splitext(image_path)
                file_extension = file_extension.lower()
                assert path.isfile(image_path), \
                    f'file {image_path} does not exist or could not be found, please ' \
                    f'check if the path of the image is valid'
            else:
                file_extension = 'base64'
        else:
            file_extension = 'BytesIO'

        assert file_extension in _VALID_IMAGE_FORMATS, \
            f'file extension {file_extension} not valid, please use: {", ".join(_VALID_IMAGE_FORMATS)}'

        self._filepath = image_path
        if isinstance(self._filepath, str) and not frombase64:
            self._filename = path.splitext(path.basename(image_path))[0]
        else:
            self._filename = ''
        self._extension = file_extension
        self._frombase64 = frombase64

        # Drawing mode
        self._drawing_mode = 0
        self._drawing_offset = (0, 0)
        self._drawing_position = ''

        self.set_drawing_mode(drawing_mode)
        self.set_drawing_offset(drawing_offset)
        self.set_drawing_position(drawing_position)

        # Convert from bas64 to bytesio
        if frombase64:
            if 'base64,' in image_path:  # Remove header of file
                for i in range(len(image_path)):
                    if image_path[i] == ',':
                        image_path = image_path[(i + 1):]
                        break
            image_path = BytesIO(base64.b64decode(image_path))

        # Load the image and store as a surface
        if load_from_file:
            self._surface = load_pygame_image_file(image_path)
            self._original_surface = self._surface.copy()

        # Other internals
        self._angle = 0
        self._last_transform = (0, 0, None)  # Improves drawing
        self._rotated = False
        self.smooth_scaling = True  # Uses smooth scaling by default in draw() method

    def __copy__(self) -> 'BaseImage':
        """
        Copy method.

        :return: New instance of the object
        """
        return self.copy()

    def __deepcopy__(self, memodict: Dict) -> 'BaseImage':
        """
        Deep-copy method.

        :param memodict: Memo dict
        :return: New instance of the object
        """
        return self.copy()

    def crop_rect(self, rect: 'pygame.Rect') -> 'BaseImage':
        """
        Crop image from rect.

        :param rect: Crop rect geometry
        :return: Self reference
        """
        self._surface = self.get_crop_rect(rect)
        return self

    def set_alpha(self, value: Optional[int], flags: int = 0) -> 'BaseImage':
        """
        Set the current alpha value for the Surface. When blitting this Surface
        onto a destination, the pixels will be drawn slightly transparent. The alpha
        value is an integer from ``0`` to ``255``, ``0`` is fully transparent and
        ``255`` is fully opaque. If None is passed for the alpha value, then alpha
        blending will be disabled, including per-pixel alpha.

        This value is different from the per pixel Surface alpha. For a surface with
        per pixel alpha, blanket alpha is ignored and None is returned.

        For pygame 2.0: per-surface alpha can be combined with per-pixel alpha.

        The optional flags argument can be set to pygame.RLEACCEL to provide better
        performance on non accelerated displays. A RLEACCEL Surface will be slower
        to modify, but quicker to blit as a source.

        :param value: Transparency value from ``0`` to ``255``
        :param flags: Optional flags
        :return: Self reference
        """
        if value is None:
            self._surface.set_alpha(None)
            return self
        assert isinstance(value, int)
        assert 0 <= value <= 255, 'alpha value must be an integer between 0 and 255'
        self._surface.set_alpha(value, flags)
        return self

    def crop(
        self,
        x: NumberType,
        y: NumberType,
        width: NumberType,
        height: NumberType
    ) -> 'BaseImage':
        """
        Crops the image from coordinate on x-axis and y-axis (x, y).

        :param x: X position within the image in px
        :param y: Y position in px
        :param width: Crop width in px
        :param height: Crop height in px
        :return: Self reference
        """
        self._surface = self.get_crop(x, y, width, height)
        return self

    def get_crop_rect(self, rect: 'pygame.Rect') -> 'pygame.Surface':
        """
        Get a crop surface of the image from rect.

        :param rect: Crop rect geometry
        :return: Cropped surface
        """
        return self._surface.subsurface(rect)

    def get_crop(
        self,
        x: NumberType,
        y: NumberType,
        width: NumberType,
        height: NumberType
    ) -> 'pygame.Surface':
        """
        Get a crop of the image from coordinate on x-axis and y-axis (x, y).

        :param x: X position within the image in px
        :param y: Y position in px
        :param width: Crop width in px
        :param height: Crop height in px
        :return: Cropped surface
        """
        assert 0 <= x < self.get_width(), \
            'X position must be between 0 and the image width'
        assert 0 <= y < self.get_height(), \
            'Y position must be between 0 and the image width'
        assert 0 < width <= self.get_width(), \
            'Width must be greater than zero and less than the image width'
        assert 0 < height <= self.get_height(), \
            'Height must be greater than zero and less than the image height'
        assert (x + width) <= self.get_width(), \
            'Crop box cannot exceed image width'
        assert (y + height) <= self.get_height(), \
            'Crop box cannot exceed image height'
        rect = pygame.Rect(0, 0, 0, 0)
        rect.x = x
        rect.y = y
        rect.width = width
        rect.height = height
        return self.get_crop_rect(rect)

    def copy(self) -> 'BaseImage':
        """
        Return a copy of the image.

        :return: Image
        """
        image = BaseImage(
            image_path=self._filepath,
            drawing_mode=self._drawing_mode,
            drawing_offset=self._drawing_offset,
            load_from_file=False,
            frombase64=self._frombase64
        )
        image._angle = self._angle
        image._surface = self._surface.copy()
        image._original_surface = self._surface.copy()
        image.smooth_scaling = self.smooth_scaling
        if self._attributes is not None:
            for k in self._attributes.keys():
                image.set_attribute(k, self._attributes[k])
        return image

    def get_path(self) -> Union[str, 'BytesIO']:
        """
        Return the image path.

        :return: Image path
        """
        return self._filepath

    def get_drawing_mode(self) -> int:
        """
        Return the image drawing mode.

        :return: Image drawing mode
        """
        return self._drawing_mode

    def set_drawing_mode(self, drawing_mode: int) -> 'BaseImage':
        """
        Set the image drawing mode.

        :param drawing_mode: Drawing mode
        :return: Self reference
        """
        assert isinstance(drawing_mode, int)
        assert drawing_mode in _VALID_IMAGE_MODES, 'unknown image drawing mode'
        self._drawing_mode = drawing_mode
        return self

    def get_drawing_offset(self) -> Tuple2IntType:
        """
        Return the image drawing offset.

        :return: Image drawing offset
        """
        return self._drawing_offset

    def set_drawing_offset(self, offset: Vector2NumberType) -> 'BaseImage':
        """
        Set the image drawing offset.

        :param offset: Drawing offset tuple on x-axis and y-axis (x, y) in px
        :return: Self reference
        """
        assert_vector(offset, 2)
        self._drawing_offset = (int(offset[0]), int(offset[1]))
        return self

    def set_drawing_position(self, position: str) -> 'BaseImage':
        """
        Set the image position.

        .. note::

            See :py:mod:`pygame_menu.locals` for valid ``position`` values.

        :param position: Image position
        :return: Self reference
        """
        assert_position(position)
        self._drawing_position = position
        return self

    def get_width(self) -> int:
        """
        Return image width in px.

        :return: Image width
        """
        return int(self._surface.get_width())

    def get_height(self) -> int:
        """
        Return image height in px.

        :return: Image height
        """
        return int(self._surface.get_height())

    def subsurface(self, rect: Union[Tuple4IntType, 'pygame.Rect']) -> 'pygame.Surface':
        """
        Return a subsurface from a rect.

        :param rect: Rect
        :return: Subsurface
        """
        return self._surface.subsurface(rect)

    def get_size(self) -> Tuple2IntType:
        """
        Return the size in pixels of the image.

        :return: Image size tuple (width, height)
        """
        return self.get_width(), self.get_height()

    def get_at(
        self,
        pos: Tuple2NumberType,
        ignore_alpha: bool = False
    ) -> Union[Tuple3IntType, Tuple4IntType, 'pygame.Color']:
        """
        Get the color from a certain position in image on x-axis and y-axis (x, y).

        ``get_at`` return a copy of the RGBA Color value at the given pixel. If the
        Surface has no per pixel alpha, then the alpha value will always be ``255``
        (opaque). If the pixel position is outside the area of the Surface an
        ``IndexError`` exception will be raised.

        Getting and setting pixels one at a time is generally too slow to be used
        in a game or realtime situation. It is better to use methods which operate
        on many pixels at a time like with the blit, fill and draw methods - or by
        using pygame.surfarraypygame module for accessing surface pixel data using
        array interfaces/pygame.PixelArraypygame object for direct pixel access of
        surfaces.

        :param pos: Position on x-axis and y-axis (x, y) in px
        :param ignore_alpha: If ``True`` returns only the three main channels
        :return: Color
        """
        assert_vector(pos, 2)
        color = self._surface.get_at(pos)
        if ignore_alpha:
            return color[0], color[1], color[2]
        return color

    def set_at(self, pos: Tuple2NumberType, color: ColorInputType) -> 'BaseImage':
        """
        Set the color of pixel on x-axis and y-axis (x, y).

        :param pos: Position on x-axis and y-axis (x, y) in px
        :param color: Color
        :return: Self reference
        """
        assert_vector(pos, 2)
        self._surface.set_at(pos, assert_color(color))
        return self

    def get_bitsize(self) -> int:
        """
        Return the image bit size.

        :return: Image size
        """
        return self._surface.get_bitsize()

    def get_surface(self, new: bool = True) -> 'pygame.Surface':
        """
        Return the surface object of the image.

        :param new: Return a new surface; if ``False`` return the same object
        :return: Image surface
        """
        if new:
            return self.get_crop_rect(self.get_rect())
        return self._surface

    def get_filename(self) -> str:
        """
        Return the name of the image file.

        :return: Filename
        """
        return self._filename

    def get_extension(self) -> str:
        """
        Return the extension of the image file.

        :return: File extension
        """
        return self._extension

    def equals(self, image: 'BaseImage') -> bool:
        """
        Return ``True`` if the image is the same as the object.

        :param image: Image object
        :return: ``True`` if the image is the same (note, the surface)
        """
        assert isinstance(image, BaseImage)
        im1 = pygame.image.tostring(self._surface, 'RGBA')
        im2 = pygame.image.tostring(image._surface, 'RGBA')
        return im1 == im2

    def restore(self) -> 'BaseImage':
        """
        Restore image to the original surface.

        :return: Self reference
        """
        self._surface = self._original_surface.copy()
        return self

    def checkpoint(self) -> 'BaseImage':
        """
        Updates the original surface to the current surface.

        :return: Self reference
        """
        self._original_surface = self._surface.copy()
        return self

    def apply_image_function(
        self,
        image_function: Callable[[int, int, int, int], Tuple4IntType]
    ) -> 'BaseImage':
        """
        Apply a function to each pixel of the image. The function will receive the
        red, green, blue and alpha colors and must return the same values. The
        color pixel will be overridden by the function output.

        .. note::

            See :py:meth:`pygame_menu.baseimage.BaseImage.to_bw` method as an
            example.

        :param image_function: Color function, takes colors as ``image_function=myfunc(r,g,b,a)``. Returns the same tuple (r, g, b, a)
        :return: Self reference
        """
        w, h = self._surface.get_size()
        for x in range(w):
            for y in range(h):
                r, g, b, a = self._surface.get_at((x, y))
                r, g, b, a = image_function(r, g, b, a)
                r = int(max(0, min(r, 255)))
                g = int(max(0, min(g, 255)))
                b = int(max(0, min(b, 255)))
                a = int(max(0, min(a, 255)))
                # noinspection PyArgumentList
                self.set_at((x, y), pygame.Color(r, g, b, a))
        return self

    def to_bw(self) -> 'BaseImage':
        """
        Converts the image to black and white.

        .. note::

            This function is slow for large images.

        :return: Self reference
        """

        def bw(r: int, g: int, b: int, a: int) -> Tuple4IntType:
            """
            To black-white function.
            """
            c = int((r + g + b) / 3)
            return c, c, c, a

        return self.apply_image_function(image_function=bw)

    def pick_channels(self, channels: ChannelType) -> 'BaseImage':
        """
        Pick certain channels of the image, channels are ``"r"`` (red), ``"g"``
        (green) and ``"b"`` (blue); ``channels param`` is a list/tuple of channels
        (non-empty).

        For example, ``pick_channels(['r', 'g'])``: All channels not included on
        the list will be discarded.

        :param channels: Channels, list or tuple containing ``"r"``, ``"g"`` or ``"b"`` (all combinations are possible)
        :return: Self reference
        """
        if isinstance(channels, str):
            channels = [channels]
        assert isinstance(channels, VectorInstance)
        assert 1 <= len(channels) <= 3, 'maximum size of channels can be 3'

        w, h = self._surface.get_size()
        for x in range(w):
            for y in range(h):
                r, g, b, a = self._surface.get_at((x, y))
                if 'r' not in channels:
                    r = 0
                if 'g' not in channels:
                    g = 0
                if 'b' not in channels:
                    b = 0
                # noinspection PyArgumentList
                self._surface.set_at((x, y), pygame.Color(r, g, b, a))
        return self

    def flip(self, x: bool, y: bool) -> 'BaseImage':
        """
        This method can flip the image either vertically, horizontally, or both.
        Flipping an image is non-destructive and does not change the dimensions.

        :param x: Flip on x-axis
        :param y: Flip on y-axis
        :return: Self reference
        """
        assert isinstance(x, bool)
        assert isinstance(y, bool)
        assert (x or y), 'at least one axis should be True'
        self._surface = pygame.transform.flip(self._surface, x, y)
        return self

    def scale(
        self,
        width: NumberType,
        height: NumberType,
        smooth: bool = True
    ) -> 'BaseImage':
        """
        Scale the image to a desired width and height factor.

        .. note::

            The scale transformation is permanent, and is applied over the same
            object. Thus, if ``image.scale(2, 2).scale(2, 2)`` the final scale of
            the initial image is equal to ``image.scale(4, 4)``.

        :param width: Width scale factor
        :param height: Height scale factor
        :param smooth: Smooth scaling
        :return: Self reference
        """
        assert isinstance(width, NumberInstance)
        assert isinstance(height, NumberInstance)
        assert isinstance(smooth, bool)
        assert width > 0 and height > 0, \
            'width and height must be greater than zero'
        w, h = self.get_size()
        if width == 1 and height == 1:
            return self
        if not smooth or self._surface.get_bitsize() < 24:
            self._surface = pygame.transform.scale(self._surface, (int(w * width), int(h * height)))
        else:  # image bitsize less than 24 bits raises ValueError
            self._surface = pygame.transform.smoothscale(self._surface, (int(w * width), int(h * height)))
        return self

    def scale2x(self) -> 'BaseImage':
        """
        This will return a new image that is double the size of the original. It
        uses the AdvanceMAME Scale2X algorithm which does a "jaggy-less" scale of
        bitmap graphics.

        This really only has an effect on simple images with solid colors. On
        photographic and anti-aliased images it will look like a regular unfiltered
        scale.

        :return: Self reference
        """
        self._surface = pygame.transform.scale2x(self._surface)
        return self

    def scale4x(self) -> 'BaseImage':
        """
        Applies a x4 scale factor using scale 2x algorithm.

        :return: Self reference
        """
        return self.scale2x().scale2x()

    def resize(
        self,
        width: NumberType,
        height: NumberType,
        smooth: bool = True
    ) -> 'BaseImage':
        """
        Resize the image to a desired (width, height) size in pixels.

        :param width: New width of the image in px
        :param height: New height of the image in px
        :param smooth: Smooth scaling
        :return: Self reference
        """
        assert isinstance(width, NumberInstance)
        assert isinstance(height, NumberInstance)
        assert isinstance(smooth, bool)
        assert width > 0 and height > 0, \
            'width and height must be greater than zero'
        w, h = self.get_size()
        if w == width and h == height:
            return self
        return self.scale(width=float(width) / w, height=float(height) / h, smooth=smooth)

    def get_rect(self) -> 'pygame.Rect':
        """
        Return the :py:class:`pygame.Rect` object of the BaseImage.

        This method returns a new rectangle covering the entire surface. The
        rectangle will always start at *(0, 0)* with a same width and height size
        as the image.

        :return: Pygame rect object
        """
        return self._surface.get_rect()

    def rotate(self, angle: NumberType, auto_checkpoint: bool = True) -> 'BaseImage':
        """
        Unfiltered counterclockwise rotation. The angle argument represents degrees
        and can be any floating point value. Negative angle amounts will rotate
        clockwise.

        .. note::

            Unless rotating by 90 degree increments, the image will be padded
            larger to hold the new size. If the image has pixel alphas, the padded
            area will be transparent. Otherwise, pygame will pick a color that matches
            the image color-key or the topleft pixel value.

        .. warning::

            Image should be rotated once. If this method is called once the Class
            rotates the previously check-pointed state. If you wish to rotate the
            current image use ``checkpoint`` to update the surface. This may
            increase the image size, because the bounding rectangle of a rotated
            image is always greater than the bounding rectangle of the original
            image (except some rotations by multiples of 90 degrees). The image
            gets distort because of the multiply copies. Each rotation generates
            a small error (inaccuracy). The sum of the errors is growing and the
            images decays.

        :param angle: Rotation angle (degrees ``0-360``)
        :param auto_checkpoint: Checkpoint after first rotation to avoid rotating the same image. If multiple rotations are applied to the same surface it will increase its size very fast because of inaccuracies
        :return: Self reference
        """
        assert isinstance(angle, NumberInstance)
        if angle == self._angle:
            return self
        if not self._rotated and auto_checkpoint:
            self.checkpoint()
        if self._rotated:
            self.restore()
        self._rotated = True
        self._surface = pygame.transform.rotate(self._surface, angle)
        self._angle = angle % 360
        return self

    def get_angle(self) -> NumberType:
        """
        Return the image angle.

        :return: Angle in degrees
        """
        return self._angle

    def _get_position_delta(self) -> Tuple2IntType:
        """
        Return the delta from drawing position.

        :return: Delta positions on x-axis and y-axis (x, y) in px
        """
        rect = self.get_rect()
        if self._drawing_position == POSITION_NORTHWEST:
            return rect.topleft
        elif self._drawing_position == POSITION_NORTH:
            return rect.midtop
        elif self._drawing_position == POSITION_NORTHEAST:
            return rect.topright
        elif self._drawing_position == POSITION_WEST:
            return rect.midleft
        elif self._drawing_position == POSITION_CENTER:
            return rect.center
        elif self._drawing_position == POSITION_EAST:
            return rect.midright
        elif self._drawing_position == POSITION_SOUTHWEST:
            return rect.bottomleft
        elif self._drawing_position == POSITION_SOUTH:
            return rect.midbottom
        elif self._drawing_position == POSITION_SOUTHEAST:
            return rect.bottomright
        else:
            raise ValueError('unknown drawing position')

    def draw(
        self,
        surface: 'pygame.Surface',
        area: Optional['pygame.Rect'] = None,
        position: Tuple2IntType = (0, 0)
    ) -> 'BaseImage':
        """
        Draw the image in a given surface.

        :param surface: Pygame surface object
        :param area: Area to draw; if ``None`` the image will be drawn on entire surface
        :param position: Position to draw on x-axis and y-axis (x, y) in px
        :return: Self reference
        """
        assert isinstance(surface, pygame.Surface)
        assert isinstance(area, (pygame.Rect, type(None)))
        assert_vector(position, 2, int)

        if area is None:
            area = surface.get_rect()

        # Compute offset based on drawing offset + drawing position
        px, py = self._get_position_delta()
        if self._drawing_mode != IMAGE_MODE_SIMPLE:
            px = 0
            py = 0

        offx = self._drawing_offset[0] - px
        offy = self._drawing_offset[1] - py

        if self._drawing_mode == IMAGE_MODE_FILL:
            # Check if exists the transformed surface
            if (
                area.width == self._last_transform[0] and
                area.height == self._last_transform[1] and
                self._last_transform[2] is not None
            ):
                surf = self._last_transform[2]
            else:  # Transform scale
                if self.smooth_scaling and self._surface.get_bitsize() > 8:
                    surf = pygame.transform.smoothscale(self._surface, (area.width, area.height))
                else:
                    surf = pygame.transform.scale(self._surface, (area.width, area.height))
                self._last_transform = (area.width, area.height, surf)

            surface.blit(
                surf,
                (
                    offx + position[0],
                    offy + position[1]
                ))

        elif self._drawing_mode == IMAGE_MODE_REPEAT_X:
            w = self._surface.get_width()
            times = int(math.ceil(float(area.width) / w))
            assert times > 0, \
                'invalid size, width must be greater than zero'
            for x in range(times):
                surface.blit(
                    self._surface,
                    (
                        x * w + offx + position[0],
                        offy + position[1]
                    ),
                    area
                )

        elif self._drawing_mode == IMAGE_MODE_REPEAT_Y:
            h = self._surface.get_height()
            times = int(math.ceil(float(area.height) / h))
            assert times > 0, \
                'invalid size, height must be greater than zero'
            for y in range(times):
                surface.blit(
                    self._surface,
                    (
                        0 + offx + position[0],
                        y * h + offy + position[1]
                    ),
                    area
                )

        elif self._drawing_mode == IMAGE_MODE_REPEAT_XY:
            w, h = self._surface.get_size()
            timesx = int(math.ceil(float(area.width) / w))
            timesy = int(math.ceil(float(area.height) / h))
            assert timesx > 0 and timesy > 0, \
                'invalid size, width and height must be greater than zero'
            for x in range(timesx):
                for y in range(timesy):
                    surface.blit(
                        self._surface,
                        (
                            x * w + offx + position[0],
                            y * h + offy + position[1]
                        ),
                        area
                    )

        elif self._drawing_mode == IMAGE_MODE_CENTER:
            sw, hw = area.width, area.height  # Window
            w, h = self._surface.get_size()  # Image
            surface.blit(
                self._surface,
                (
                    int(float(sw - w) / 2 + offx + position[0]),
                    int(float(hw - h) / 2 + offy + position[1])
                ),
                area
            )

        elif self._drawing_mode == IMAGE_MODE_SIMPLE:
            surface.blit(
                self._surface,
                (
                    offx + position[0],
                    offy + position[1]
                ),
                area
            )

        return self
